from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import cv2
import numpy as np
import librosa.feature as librosa_feature
from librosa import load as librosa_load
from PIL import Image
from tensorflow.keras.preprocessing.text import Tokenizer

from terra_ai_datasets.creation.validators.inputs import ImageNetworkTypes, ImageValidator, TextValidator, \
    TextProcessTypes, TextModeTypes, AudioValidator, AudioParameterTypes, RawValidator, TimeseriesValidator, \
    ImageScalers, ImageProcessTypes
from terra_ai_datasets.creation.validators.outputs import SegmentationValidator, ClassificationValidator, \
    RegressionValidator, DepthValidator, TrendValidator


def resize_frame(image_array, target_shape, frame_mode):
    original_shape = (image_array.shape[0], image_array.shape[1])
    resized = None
    if frame_mode == ImageProcessTypes.stretch:
        resized = cv2.resize(image_array, (target_shape[1], target_shape[0]))

    elif frame_mode == ImageProcessTypes.fit:
        if image_array.shape[1] >= image_array.shape[0]:
            resized_shape = list(target_shape).copy()
            resized_shape[0] = int(
                image_array.shape[0] / (image_array.shape[1] / target_shape[1])
            )
            if resized_shape[0] > target_shape[0]:
                resized_shape = list(target_shape).copy()
                resized_shape[1] = int(
                    image_array.shape[1] / (image_array.shape[0] / target_shape[0])
                )
            image_array = cv2.resize(image_array, (resized_shape[1], resized_shape[0]))
        elif image_array.shape[0] >= image_array.shape[1]:
            resized_shape = list(target_shape).copy()
            resized_shape[1] = int(
                image_array.shape[1] / (image_array.shape[0] / target_shape[0])
            )
            if resized_shape[1] > target_shape[1]:
                resized_shape = list(target_shape).copy()
                resized_shape[0] = int(
                    image_array.shape[0] / (image_array.shape[1] / target_shape[1])
                )
            image_array = cv2.resize(image_array, (resized_shape[1], resized_shape[0]))
        resized = image_array
        if resized.shape[0] < target_shape[0]:
            black_bar = np.zeros(
                (int((target_shape[0] - resized.shape[0]) / 2), resized.shape[1], 3),
                dtype="uint8",
            )
            resized = np.concatenate((black_bar, resized))
            black_bar_2 = np.zeros(
                (int((target_shape[0] - resized.shape[0])), resized.shape[1], 3),
                dtype="uint8",
            )
            resized = np.concatenate((resized, black_bar_2))
        if resized.shape[1] < target_shape[1]:
            black_bar = np.zeros(
                (target_shape[0], int((target_shape[1] - resized.shape[1]) / 2), 3),
                dtype="uint8",
            )
            resized = np.concatenate((black_bar, resized), axis=1)
            black_bar_2 = np.zeros(
                (target_shape[0], int((target_shape[1] - resized.shape[1])), 3),
                dtype="uint8",
            )
            resized = np.concatenate((resized, black_bar_2), axis=1)

    elif frame_mode == ImageProcessTypes.cut:
        resized = image_array.copy()
        if original_shape[0] > target_shape[0]:
            resized = resized[
                int(original_shape[0] / 2 - target_shape[0] / 2) : int(
                    original_shape[0] / 2 - target_shape[0] / 2
                )
                + target_shape[0],
                :,
            ]
        else:
            black_bar = np.zeros(
                (int((target_shape[0] - original_shape[0]) / 2), original_shape[1], 3),
                dtype="uint8",
            )
            resized = np.concatenate((black_bar, resized))
            black_bar_2 = np.zeros(
                (int((target_shape[0] - resized.shape[0])), original_shape[1], 3),
                dtype="uint8",
            )
            resized = np.concatenate((resized, black_bar_2))
        if original_shape[1] > target_shape[1]:
            resized = resized[
                :,
                int(original_shape[1] / 2 - target_shape[1] / 2) : int(
                    original_shape[1] / 2 - target_shape[1] / 2
                )
                + target_shape[1],
            ]
        else:
            black_bar = np.zeros(
                (target_shape[0], int((target_shape[1] - original_shape[1]) / 2), 3),
                dtype="uint8",
            )
            resized = np.concatenate((black_bar, resized), axis=1)
            black_bar_2 = np.zeros(
                (target_shape[0], int((target_shape[1] - resized.shape[1])), 3),
                dtype="uint8",
            )
            resized = np.concatenate((resized, black_bar_2), axis=1)
    return resized


class Array(ABC):

    @abstractmethod
    def create(self, source: Any, parameters: Any):
        pass

    @abstractmethod
    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: Any):
        pass


class ImageArray(Array):

    def create(self, source: Path, parameters: ImageValidator):

        image = Image.open(str(source).split(";")[0])
        array = np.asarray(image)
        if array.ndim == 2:
            array = np.stack([array, array, array], axis=-1)
        array = resize_frame(image_array=array,
                             target_shape=(parameters.height, parameters.width),
                             frame_mode=parameters.process)
        if parameters.network == ImageNetworkTypes.linear:
            array = array.reshape(np.prod(np.array(array.shape)))

        return array

    def preprocess(self, array: np.ndarray, preprocess_obj: Any, parameters: ImageValidator) -> np.ndarray:
        if parameters.preprocessing == ImageScalers.terra_image_scaler:
            array = array.astype(np.float32)
            if len(array.shape) > 3:
                for i in range(len(array)):
                    array[i] = preprocess_obj.transform(array[i])
            else:
                array = preprocess_obj.transform(array)
        else:
            orig_shape = array.shape
            array = preprocess_obj.transform(array.reshape(-1, 1)).astype(np.float32)
            array = array.reshape(orig_shape)

        return array


class TextArray(Array):

    def create(self, source: str, parameters: TextValidator):

        return source

    def preprocess(self, text_list: list, preprocess_obj: Tokenizer, parameters: TextValidator) -> np.ndarray:

        array = []
        if not type(text_list) == np.ndarray:
            text_list = [text_list]
        for text in text_list:
            if parameters.preprocessing == TextProcessTypes.embedding:
                text_array = preprocess_obj.texts_to_sequences([text])[0]
                if parameters.mode == TextModeTypes.full and len(text_array) < parameters.max_words:
                    text_array += [0 for _ in range(parameters.max_words - len(text_array))]
                elif parameters.mode == TextModeTypes.full and len(text_array) > parameters.max_words:
                    text_array = text_array[:parameters.max_words]
                elif parameters.mode == TextModeTypes.length_and_step and len(text_array) < parameters.length:
                    text_array += [0 for _ in range(parameters.length - len(text_array))]
            elif parameters.preprocessing == TextProcessTypes.bag_of_words:
                text_array = preprocess_obj.texts_to_matrix([text])[0].astype(np.int8)
            elif parameters.preprocessing == TextProcessTypes.word_to_vec:
                text_array = []
                for word in text.split(' '):
                    try:
                        text_array.append(preprocess_obj.wv[word])
                    except KeyError:
                        text_array.append(np.zeros((parameters.word2vec_size,)))
                if len(text_array) < parameters.length:
                    words_to_add = [[0 for _ in range(parameters.word2vec_size)]
                                    for _ in range(parameters.length - len(text_array))]
                    text_array += words_to_add
            array.append(text_array)

        return np.array(array)


class AudioArray(Array):

    def create(self, source: str, parameters: AudioValidator):
        array = []
        parameter = parameters.parameter[0]
        audio_path, start_stop = source.split(';')
        offset, stop = [float(x) for x in start_stop.split(':')]
        duration = stop - offset

        y, sr = librosa_load(
            path=audio_path, sr=parameters.sample_rate, offset=offset, duration=duration,
            res_type=parameters.resample.name
        )

        if round(parameters.sample_rate * duration, 0) > y.shape[0]:
            zeros = np.zeros((int(round(parameters.sample_rate * duration, 0)) - y.shape[0],))
            y = np.concatenate((y, zeros))
        if parameter in [AudioParameterTypes.chroma_stft, AudioParameterTypes.mfcc,
                         AudioParameterTypes.spectral_centroid, AudioParameterTypes.spectral_bandwidth,
                         AudioParameterTypes.spectral_rolloff]:
            array = getattr(librosa_feature, parameter.name)(y=y, sr=parameters.sample_rate)
        elif parameter == AudioParameterTypes.rms:
            array = getattr(librosa_feature, parameter.name)(y=y)[0]
        elif parameter == AudioParameterTypes.zero_crossing_rate:
            array = getattr(librosa_feature, parameter.name)(y=y)
        elif parameter == AudioParameterTypes.audio_signal:
            array = y

        array = np.array(array)
        if len(array.shape) == 2:
            array = array.transpose()
        if array.dtype == 'float64':
            array = array.astype('float32')

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: Any):

        orig_shape = array.shape
        if len(orig_shape) > 1:
            array = array.reshape(-1, 1)
        array = preprocess.transform(array)
        array = array.reshape(orig_shape)

        return array


class RawArray(Array):

    def create(self, source: str, parameters: RawValidator):

        array = np.array([[source]])

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: RawValidator):

        return array


class TimeseriesArray(Array):

    def create(self, source: list, parameters: TimeseriesValidator):

        array = np.array(source)
        array = np.expand_dims(array, 1)

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: TimeseriesValidator):

        orig_shape = array.shape
        array = array.reshape(-1, 1)
        array = preprocess.transform(array)
        array = array.reshape(orig_shape)

        return array


class ClassificationArray(Array):

    def create(self, source: str, parameters: ClassificationValidator):

        array = np.array([parameters.classes_names.index(source)], dtype=np.uint8)
        if parameters.one_hot_encoding:
            zeros = np.zeros(len(parameters.classes_names), dtype=np.uint8)
            zeros[array[0]] = 1
            array = zeros

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: ClassificationValidator):

        return array


class CategoricalArray(ClassificationArray):
    pass


class SegmentationArray(Array):

    def create(self, source: str, parameters: SegmentationValidator):

        image = Image.open(source)
        array = np.asarray(image)
        if array.ndim == 2:
            array = np.stack([array, array, array], axis=-1)
        array = resize_frame(image_array=array,
                             target_shape=(parameters.height, parameters.width),
                             frame_mode=parameters.process)
        array = self.image_to_ohe(array, parameters)

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: SegmentationValidator):

        return array

    @staticmethod
    def image_to_ohe(img_array, parameters: SegmentationValidator):
        mask_ohe = []
        mask_range = parameters.rgb_range
        for color_obj in parameters.classes.values():
            color = color_obj.as_rgb_tuple()
            color_array = np.expand_dims(np.where((color[0] + mask_range >= img_array[:, :, 0]) &
                                                  (img_array[:, :, 0] >= color[0] - mask_range) &
                                                  (color[1] + mask_range >= img_array[:, :, 1]) &
                                                  (img_array[:, :, 1] >= color[1] - mask_range) &
                                                  (color[2] + mask_range >= img_array[:, :, 2]) &
                                                  (img_array[:, :, 2] >= color[2] - mask_range), 1, 0),
                                         axis=2)
            mask_ohe.append(color_array)

        return np.concatenate(np.array(mask_ohe), axis=2).astype(np.uint8)


class RegressionArray(Array):

    def create(self, source: str, parameters: RegressionValidator):

        array = np.array([float(source)])

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: RegressionValidator):

        orig_shape = array.shape
        array = array.reshape(-1, 1)
        array = preprocess.transform(array)
        array = array.reshape(orig_shape)

        return array


class DepthArray(Array):

    def create(self, source: list, parameters: DepthValidator):

        array = np.array(source)[:parameters.depth]
        array = np.expand_dims(array, 1)

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: DepthValidator):

        orig_shape = array.shape
        array = array.reshape(-1, 1)
        array = preprocess.transform(array)
        array = array.reshape(orig_shape)

        return array


class TrendArray(Array):

    def create(self, source: list, parameters: TrendValidator):

        source = np.array(source)
        first_val, second_val = source[0], source[1]
        if abs((second_val - first_val) / first_val) * 100 <= parameters.deviation:
            array = 0
        elif second_val > first_val:
            array = 1
        else:
            array = 2
        if parameters.one_hot_encoding:
            ohe_array = np.zeros((3,))
            ohe_array[array] = 1
            array = ohe_array.astype(np.uint8)

        return array

    def preprocess(self, array: np.ndarray, preprocess: Any, parameters: TrendValidator):

        return array



from enum import Enum
from typing import Tuple, Dict

from pydantic import BaseModel, validator
from pydantic.types import PositiveInt


class AugmentationMode(Enum):
    light = "Light"
    medium = "Medium"
    maximum = "Maximum"


# --- Image augmentation ---
class ImageAugmentationFliplr(BaseModel):
    p: float

    @validator("p")
    def _validate_p(cls, v):
        if not 0.0 <= v <= 1.0:
            raise ValueError('Ensure "p" value is in range between 0.0 and 1.0')
        return v


class ImageAugmentationFlipud(BaseModel):
    p: float

    @validator("p")
    def _validate_p(cls, v):
        if not 0.0 <= v <= 1.0:
            raise ValueError('Ensure "p" value is in range between 0.0 and 1.0')
        return v


class ImageAugmentationCrop(BaseModel):
    px: Tuple[PositiveInt, PositiveInt]


class ImageAugmentationLinearContrast(BaseModel):
    alpha: Tuple[float, float]

    @validator("alpha")
    def _validate_alpha(cls, v):
        if not v[0] < v[1]:
            raise ValueError(f'First value "{v[0]}" must be no greater than second "{v[1]}"')
        return v


class ImageAugmentationAffine(BaseModel):
    scale: Dict[str, Tuple[float, float]]
    translate_percent: Dict[str, Tuple[float, float]]
    rotate: Tuple[int, int]
    shear: Tuple[int, int]


class ImageAugmentationAdditiveGaussianNoise(BaseModel):
    loc: int
    scale: Tuple[float, float]
    per_channel: float


class ImageAugmentationMultiply(BaseModel):
    mul: Tuple[float, float]
    per_channel: float


class ImageAugmentationParameters(BaseModel):
    Fliplr: ImageAugmentationFliplr = None
    Flipud: ImageAugmentationFlipud = None
    Crop: ImageAugmentationCrop = None
    LinearContrast: ImageAugmentationLinearContrast = None
    AdditiveGaussianNoise: ImageAugmentationAdditiveGaussianNoise = None
    Multiply: ImageAugmentationMultiply = None
    Affine: ImageAugmentationAffine = None


image_mode_config = {
    AugmentationMode.light.value: {
        "Fliplr": {"p": 0.5},
        "Crop": {"px": (1, 16)},
    },
    AugmentationMode.medium.value: {
        "Fliplr": {"p": 0.5},
        "Flipud": {"p": 0.5},
        "Crop": {"px": (1, 16)},
        "LinearContrast": {"alpha": (0.75, 1.5)},
        "AdditiveGaussianNoise": {"loc": 0, "scale": (0.0, 0.05 * 255), "per_channel": 0.5}
    },
    AugmentationMode.maximum.value: {
        "Fliplr": {"p": 0.5},
        "Flipud": {"p": 0.5},
        "Crop": {"px": (1, 16)},
        "LinearContrast": {"alpha": (0.75, 1.5)},
        "Multiply": {"mul": (0.8, 1.2), "per_channel": 0.2},
        "AdditiveGaussianNoise": {"loc": 0, "scale": (0.0, 0.05 * 255), "per_channel": 0.5},
        "Affine": {
            "scale": {"x": (0.8, 1.2), "y": (0.8, 1.2)},
            "translate_percent": {"x": (-0.2, 0.2), "y": (-0.2, 0.2)},
            "rotate": (-25, 25),
            "shear": (-8, 8)
        }
    }
}



import logging
import os
import re
from enum import Enum
from typing import Union, List, Any

import librosa
import numpy as np
import pandas as pd
import pymorphy2
from pathlib import Path
from PIL import Image

from sklearn.cluster import KMeans
from tensorflow.keras.preprocessing.text import text_to_word_sequence

from terra_ai_datasets.creation import arrays
from terra_ai_datasets.creation.arrays import resize_frame
from terra_ai_datasets.creation.validators.creation_data import InputInstructionsData, OutputInstructionsData
from terra_ai_datasets.creation.validators.inputs import TextValidator, TextModeTypes, ImageValidator, AudioValidator
from terra_ai_datasets.creation.validators.outputs import SegmentationValidator

logger_formatter = logging.Formatter(f"%(asctime)s | %(message)s", "%H:%M:%S")
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(logger_formatter)

logger = logging.getLogger(__name__)
if logger.hasHandlers():
    logger.handlers.clear()
logger.addHandler(stream_handler)
logger.setLevel("INFO")

ANNOTATION_SEPARATOR = ":"
ANNOTATION_LABEL_NAME = "# label"
ANNOTATION_COLOR_RGB = "color_rgb"


class DatasetFoldersData(Enum):
    arrays = "arrays"
    instructions = "instructions"
    preprocessing = "preprocessing"
    sources = "sources"


class DatasetPutsData(Enum):
    inputs = "inputs"
    outputs = "outputs"


class DatasetInstructionFoldersData(Enum):
    dataframe = "dataframe"
    parameters = "parameters"


class DatasetFoldersPutsData:

    def __init__(self, native_path: Path):
        for folder in DatasetPutsData:
            native_path.joinpath(folder.value).mkdir(parents=True, exist_ok=True)
        self.native = native_path

    @property
    def inputs(self):
        return self.native.joinpath(DatasetPutsData.inputs.value)

    @property
    def outputs(self):
        return self.native.joinpath(DatasetPutsData.outputs.value)


class DatasetInstructionsFoldersData:

    def __init__(self, native_path: Path):
        for folder in DatasetInstructionFoldersData:
            native_path.joinpath(folder.value).mkdir(parents=True, exist_ok=True)
        self.native = native_path

    @property
    def dataframe(self):
        return self.native.joinpath(DatasetInstructionFoldersData.dataframe.value)

    @property
    def parameters(self):
        return self.native.joinpath(DatasetInstructionFoldersData.parameters.value)


class DatasetPathsData:

    def __init__(self, root_path: Path):
        self.root_path = root_path
        self.arrays = DatasetFoldersPutsData(root_path.joinpath(DatasetFoldersData.arrays.value))
        self.instructions = DatasetInstructionsFoldersData(root_path.joinpath(DatasetFoldersData.instructions.value))
        self.preprocessing = root_path.joinpath(DatasetFoldersData.preprocessing.value)
        self.sources = root_path.joinpath(DatasetFoldersData.sources.value)
        self.preprocessing.mkdir(exist_ok=True)
        self.sources.mkdir(exist_ok=True)

    @property
    def config(self):
        return self.root_path.joinpath("config.json")


def decamelize(camel_case_string: str):
    string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", camel_case_string)
    return re.sub("([a-z0-9])([A-Z0-9])", r"\1_\2", string).lower()


def create_put_array(data: Any, put_data: Union[InputInstructionsData, OutputInstructionsData]):
    sample_array = getattr(arrays, f"{put_data.type}Array")().create(data, put_data.parameters)

    return sample_array


def preprocess_put_array(
        data: Any, put_data: Union[InputInstructionsData, OutputInstructionsData], preprocessing: Any
):

    preprocessed_array = getattr(arrays, f"{put_data.type}Array")().preprocess(
        data, preprocessing, put_data.parameters
    )

    return preprocessed_array


def process_directory_paths(dir_paths: list) -> List[Path]:

    list_of_directories = []
    for dir_path in dir_paths:
        for d_path, dir_names, _ in os.walk(dir_path):
            if not dir_names:
                list_of_directories.append(Path(d_path))

    return list_of_directories


def apply_pymorphy(data_to_process: list):
    pymorphy = pymorphy2.MorphAnalyzer()

    data_to_return = []
    for text in data_to_process:
        words_list = text.split(' ')
        words_list = [pymorphy.parse(w)[0].normal_form for w in words_list]
        data_to_return.append(' '.join(words_list))

    return data_to_return


def extract_image_data(folder_path: Path, parameters: Union[ImageValidator, SegmentationValidator]):
    image_samples = []
    for image_path in sorted(folder_path.iterdir()):
        image_samples.append(image_path)

    return image_samples


def extract_segmentation_data(folder_path: Path, parameters: SegmentationValidator):
    return extract_image_data(folder_path, parameters)


def extract_text_data(folder_path: Path, parameters: TextValidator):
    text_samples = []
    for text_path in sorted(folder_path.iterdir()):
        with open(text_path, 'r', encoding="utf-8") as text_file:
            text = ' '.join(text_to_word_sequence(
                text_file.read().strip(), **{'lower': False, 'filters': parameters.filters, 'split': ' '})
            ).split()
        if parameters.mode == TextModeTypes.full:
            text_samples.append(' '.join(text[:parameters.max_words]))
        else:
            for i in range(0, len(text), parameters.step):
                text_sample = text[i: i + parameters.length]
                text_samples.append(' '.join(text_sample))
                if len(text_sample) < parameters.length:
                    break
    return text_samples


def extract_audio_data(folder_path: Path, parameters: AudioValidator):
    audio_samples = []
    for audio_path in sorted(folder_path.iterdir()):
        if parameters.mode == TextModeTypes.full:
            audio_samples.append(';'.join([str(audio_path), f"0:{parameters.max_seconds}"]))
        else:
            duration = librosa.get_duration(path=str(audio_path))
            start_idx, stop_idx = 0, parameters.length
            audio_samples.append(';'.join([str(audio_path), f"{start_idx}:{stop_idx}"]))
            while stop_idx < duration:
                start_idx += parameters.step
                stop_idx += parameters.step
                audio_samples.append(';'.join([str(audio_path), f"{start_idx}:{stop_idx}"]))
    return audio_samples


def get_classes_autosearch(source, num_classes: int, mask_range: int, height, width, frame_mode):

    def _rgb_in_range(rgb: tuple, target: tuple) -> bool:
        _range0 = range(target[0] - mask_range, target[0] + mask_range)
        _range1 = range(target[1] - mask_range, target[1] + mask_range)
        _range2 = range(target[2] - mask_range, target[2] + mask_range)
        return rgb[0] in _range0 and rgb[1] in _range1 and rgb[2] in _range2

    source = process_directory_paths(source)

    annotations = {}
    for dirname in source:
        for filepath in dirname.iterdir():
            if len(annotations) >= num_classes:
                break

            array = np.asarray(Image.open(filepath))
            array = resize_frame(image_array=array, target_shape=(height, width), frame_mode=frame_mode)

            np_data = array.reshape(-1, 3)
            km = KMeans(n_clusters=num_classes)
            km.fit(np_data)

            cluster_centers = (
                np.round(km.cluster_centers_)
                .astype("uint8")[: max(km.labels_) + 1]
                .tolist()
            )

            for index, rgb in enumerate(cluster_centers, 1):
                if rgb in annotations.values():
                    continue

                add_condition = True
                for rgb_target in annotations.values():
                    if _rgb_in_range(tuple(rgb), rgb_target):
                        add_condition = False
                        break

                if add_condition:
                    annotations[str(index)] = rgb

    return annotations


def get_classes_annotation(path: str):

    path = Path(path)
    assert path.is_file(), 'Путь до файла аннотации указан неверно'

    data = pd.read_csv(path, sep=ANNOTATION_SEPARATOR)

    annotations = {}
    for row in data.iterrows():
        index, row_data = row
        annotations[getattr(row_data, ANNOTATION_LABEL_NAME)] = \
            [int(num) for num in getattr(row_data, ANNOTATION_COLOR_RGB).split(',')]

    return annotations



from typing import Dict, Union, Any

import numpy as np
import pandas as pd
from imgaug import SegmentationMapsOnImage
from tqdm import tqdm

from terra_ai_datasets.creation.augmentation import create_image_augmentation
from terra_ai_datasets.creation.dataset import CreateDataset, CreateClassificationDataset
from terra_ai_datasets.creation.utils import get_classes_annotation, get_classes_autosearch, logger
from terra_ai_datasets.creation.validators.creation_data import InputData, OutputData, InputInstructionsData, \
    OutputInstructionsData
from terra_ai_datasets.creation.validators import inputs as val_inputs
from terra_ai_datasets.creation.validators import outputs as val_outputs
from terra_ai_datasets.creation.validators.inputs import ImageScalers, TextModeTypes, TextProcessTypes
from terra_ai_datasets.creation.validators.tasks import LayerInputTypeChoice, LayerOutputTypeChoice, \
    LayerSelectTypeChoice


class ImageClassification(CreateClassificationDataset):
    input_type = LayerInputTypeChoice.Image
    output_type = LayerOutputTypeChoice.Classification

    def __init__(
            self,
            source_path: list,
            train_size: float,
            width: int,
            height: int,
            network: str,
            process: str,
            preprocessing: str = ImageScalers.none,
            one_hot_encoding: bool = True,
            augmentation: str = None,
            augmentation_coef: float = 0
    ):
        """
        Класс подготовки датасета для задачи классификации изображений.

        :param source_path: Список относительных путей до папок с изображениями;
        :param train_size: Соотношение обучающей выборки к валидационной;
        :param width: Ширина изображений;
        :param height: Высота изображений;
        :param process: Метод обработки изображений при изменении размерности. Варианты: "Stretch", "Fit", "Cut";
        :param network: Постобработка массивов под определенный вид нейронной сети. Варианты: "Convolutional", "Linear";
        :param preprocessing: Выбор скейлера. Варианты: "None", "MinMaxScaler", "TerraImageScaler";
        :param one_hot_encoding: Перевод Y массивов в формат One-Hot Encoding.
        :param augmentation: Словарь с параметрами для аугментации изображений.
        """
        super().__init__(source_path=source_path, train_size=train_size, width=width, height=height, process=process,
                         network=network, preprocessing=preprocessing, one_hot_encoding=one_hot_encoding,
                         augmentation=augmentation, augmentation_coef=augmentation_coef)


class ImageSegmentation(CreateDataset):
    input_type = LayerInputTypeChoice.Image
    output_type = LayerOutputTypeChoice.Segmentation

    def __init__(
            self,
            source_path: list,
            target_path: list,
            train_size: float,
            width: int,
            height: int,
            network: str,
            process: str,
            rgb_range: int,
            classes: Dict[str, list] = None,
            num_classes: int = None,
            classes_path: str = None,
            preprocessing: str = ImageScalers.none,
            augmentation: str = None,
            augmentation_coef: float = 0
    ):
        """
        Класс подготовки датасета для задачи сегментации изображений.

        :param source_path: Список относительных путей до папок с изображениями;
        :param target_path: Список относительных путей до папок с масками сегментации;
        :param train_size: Соотношение обучающей выборки к валидационной;
        :param width: Ширина изображений;
        :param height: Высота изображений;
        :param process: Метод обработки изображений при изменении размерности. Варианты: "Stretch", "Fit", "Cut";
        :param network: Постобработка массивов под определенный вид нейронной сети. Варианты: "Convolutional", "Linear";
        :param preprocessing: Выбор скейлера. Варианты: "None", "MinMaxScaler", "TerraImageScaler";
        :param rgb_range: Диапазон, при котором пиксели будут отнесены к классу;
        :param classes: Названия классов и их RGB значения в виде словаря;
        :param classes_path: Путь к txt файлу с указанием названия классов и их RGB значений.
        :param augmentation: Словарь с параметрами для аугментации изображений.
        """
        if not classes and not classes_path and num_classes:
            classes = get_classes_autosearch(source=target_path, num_classes=num_classes, mask_range=rgb_range,
                                             height=height, width=width, frame_mode=process)
        elif not classes and classes_path:
            classes = get_classes_annotation(classes_path)

        super().__init__(source_path=source_path, target_path=target_path, train_size=train_size,  width=width,
                         height=height, process=process, network=network, preprocessing=preprocessing,
                         rgb_range=rgb_range, classes=classes, augmentation=augmentation,
                         augmentation_coef=augmentation_coef)

    def preprocess_put_data(self, data, data_type: LayerSelectTypeChoice):
        puts_data = super().preprocess_put_data(data, data_type)

        puts_data[2][f"2_{self.output_type.value}"].parameters.height = \
            puts_data[1][f"1_{self.input_type.value}"].parameters.height
        puts_data[2][f"2_{self.output_type.value}"].parameters.width = \
            puts_data[1][f"1_{self.input_type.value}"].parameters.width
        puts_data[2][f"2_{self.output_type.value}"].parameters.process = \
            puts_data[1][f"1_{self.input_type.value}"].parameters.process

        return puts_data

    def create(self, use_generator: bool = False, verbose: int = 1):
        super().create(use_generator=use_generator, verbose=verbose)
        if not use_generator:
            if verbose == 0:
                logger.info(f"Выполняется аугментация изображений и масок сегментации")

            aug_idx = self.dataframe["train"]["1_Image"].str.contains("augm").tolist()
            orig_shape = self.X['train']['input_1'][0].shape
            for i, val in enumerate(aug_idx):
                if val == True:  # Не менять!
                    x_sample = self.X['train']['input_1'][i]
                    y_sample = self.Y['train']['output_1'][i]
                    if self.preprocessing.get("1_Image"):
                        x_sample = self.preprocessing["1_Image"].inverse_transform(x_sample.reshape(-1, 1)).reshape(orig_shape).astype(np.uint8)
                    y_maps = SegmentationMapsOnImage(y_sample, shape=y_sample.shape)
                    x_sample, y_sample = self.augmentation["1_Image"](image=x_sample, segmentation_maps=y_maps)
                    if self.preprocessing.get("1_Image"):
                        x_sample = self.preprocessing["1_Image"].transform(x_sample.reshape(-1, 1)).reshape(orig_shape)
                    self.X["train"]["input_1"][i] = x_sample
                    self.Y["train"]["output_1"][i] = y_sample.arr

            if verbose == 0:
                logger.info(f"Готово!")


class TextClassification(CreateClassificationDataset):
    input_type = LayerInputTypeChoice.Text
    output_type = LayerOutputTypeChoice.Classification

    def __init__(
            self,
            source_path: list,
            train_size: float,
            preprocessing: str = TextProcessTypes.embedding,
            max_words_count: int = None,
            word2vec_size: int = None,
            mode: str = TextModeTypes.full,
            filters: str = '–—!"#$%&()*+,-./:;<=>?@[\\]^«»№_`{|}~\t\n\xa0–\ufeff',
            max_words: int = None,
            length: int = None,
            step: int = None,
            pymorphy: bool = False,
            one_hot_encoding: bool = True
    ):
        """
        Класс подготовки датасета для задачи классификации текстов.

        :param source_path: Список относительных путей до папок с изображениями;
        :param train_size: Соотношение обучающей выборки к валидационной;
        :param preprocessing: Тип обработки текстов. Варианты: "Embedding", "Bag of words", "Word2Vec";
        :param max_words_count: Максимальное количество слов, сохраняемое в словаре частотности;
        :param mode: Режим обрезки текстовых файлов. Варианты: "Full", "Length and step";
        :param filters: Символы, подлежащие удалению;
        :param pymorphy: Применить морфологический анализатор (приведение слов в нормальную форму);
        :param max_words: ТОЛЬКО при mode == "Full" - Максимальное количество слов, извлекаемое из одного файла;
        :param length: ТОЛЬКО при mode == "Length and step" - Количество слов в одном примере;
        :param step: ТОЛЬКО при mode == "Length and step" - Шаг движения по тексту при составлении примеров;
        :param one_hot_encoding: Перевод Y массивов в формат One-Hot Encoding.
        """
        parameters = {"source_path": source_path, "train_size": train_size, "preprocessing": preprocessing,
                      "max_words_count": max_words_count, "mode": mode, "filters": filters, "pymorphy": pymorphy,
                      "one_hot_encoding": one_hot_encoding}

        for name, param in {"max_words": max_words, "length": length, "step": step,
                            "word2vec_size": word2vec_size}.items():
            if param:
                parameters[name] = param

        super().__init__(**parameters)


class AudioClassification(CreateClassificationDataset):
    input_type = LayerInputTypeChoice.Audio
    output_type = LayerOutputTypeChoice.Classification

    def __init__(
            self,
            source_path: list,
            train_size: float,
            mode: str,
            parameter: list,
            fill_mode: str,
            resample: str,
            max_seconds: float = None,
            length: float = None,
            step: float = None,
            sample_rate: int = 22050,
            preprocessing: str = 'None',
            one_hot_encoding: bool = True
    ):
        """
        Класс подготовки датасета для задачи классификации аудио.

        :param source_path: Список относительных путей до папок с изображениями;
        :param train_size: Соотношение обучающей выборки к валидационной;
        :param mode: Режим обрезки аудио файлов. Варианты: "Full", "Length and step";
        :param parameter: Спсиок извлекаемых фичей из аудио файлов;
        :param fill_mode: Режим заполнения при недостатке длины. Варианты: "Loop", "Last millisecond";
        :param resample: Режим ресемпла при открытии аудио файла. Варианты: "Kaiser best", "Kaiser fast", "Scipy";
        :param max_seconds: ТОЛЬКО при mode == "Full" - Максимальное количество секунд, извлекаемое из одного файла;
        :param length: ТОЛЬКО при mode == "Length and step" - Длина одного примера;
        :param step: ТОЛЬКО при mode == "Length and step" - Шаг движения по аудиофайлу при составлении примеров;
        :param sample_rate: Sample rate при открытии аудиофайла;
        :param preprocessing: Выбор скейлера. Варианты: "StandardScaler", "MinMaxScaler";
        :param one_hot_encoding: Перевод Y массивов в формат One-Hot Encoding.
        """
        parameters = {"source_path": source_path, "train_size": train_size, "preprocessing": preprocessing,
                      "parameter": parameter, "mode": mode, "sample_rate": sample_rate, "fill_mode": fill_mode,
                      "resample": resample, "one_hot_encoding": one_hot_encoding}

        for name, param in {"max_seconds": max_seconds, "length": length, "step": step}.items():
            if param:
                parameters[name] = param

        super().__init__(**parameters)

    def preprocess_put_data(self, data, data_type: LayerSelectTypeChoice):
        put_data = super().preprocess_put_data(data, data_type)

        input_data = put_data[1][f"1_{self.input_type.value}"]
        new_put_data = {}

        for put_id, parameter in enumerate(input_data.parameters.parameter, 1):
            new_params_data = input_data.parameters.dict()
            new_params_data['parameter'] = [parameter]
            new_input_data = InputData(
                folder_path=input_data.folder_path,
                type=self.input_type,
                parameters=getattr(val_inputs, f"{self.input_type.value}Validator")(**new_params_data)
            )
            new_put_data[put_id] = {f"{put_id}_Audio": new_input_data}

        new_put_data[len(new_put_data) + 1] = \
            {f"{len(new_put_data) + 1}_{self.output_type.value}": put_data[2][f"2_{self.output_type.value}"]}

        return new_put_data


class Dataframe(CreateDataset):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)


class DataframeRegression(Dataframe):
    input_type = LayerInputTypeChoice.Dataframe
    output_type = LayerOutputTypeChoice.Regression

    def __init__(
            self,
            csv_path: str,
            train_size: float,
            inputs: list,
            output: str,
            preprocessing: str
    ):
        super().__init__(csv_path=csv_path, inputs=inputs, output=output,  train_size=train_size,
                         preprocessing=preprocessing)


class DataframeClassification(Dataframe):
    input_type = LayerInputTypeChoice.Dataframe
    output_type = LayerOutputTypeChoice.Classification

    def __init__(
            self,
            csv_path: str,
            train_size: float,
            inputs: list,
            output: str,
            one_hot_encoding: bool,
    ):
        super().__init__(csv_path=csv_path, inputs=inputs, output=output,  train_size=train_size,
                         one_hot_encoding=one_hot_encoding)


class TimeseriesDepth(Dataframe):
    input_type = LayerInputTypeChoice.Timeseries
    output_type = LayerOutputTypeChoice.Depth

    def __init__(
            self,
            csv_path: str,
            train_size: float,
            inputs: list,
            outputs: list,
            preprocessing: str,
            length: int,
            step: int,
            depth: int,
    ):
        """
        Класс подготовки датасета для задачи временных рядов.

        :param csv_path: Путь до csv-файла;
        :param train_size: Соотношение обучающей выборки к валидационной;
        :param inputs: Список колонок, используемых в качестве входа;
        :param outputs: Список колонок, используемых в качестве выхода;
        :param preprocessing: Выбор скейлера. Варианты: "StandardScaler", "MinMaxScaler";
        :param length: Длина одного примера;
        :param step: Длина шага окна разбивки при составлении примеров;
        :param depth: Глубина предсказания;
        """
        super().__init__(csv_path=csv_path, inputs=inputs, outputs=outputs,  train_size=train_size,
                         preprocessing=preprocessing, length=length, step=step, depth=depth)

    def preprocess_put_data(self, data, data_type: LayerSelectTypeChoice) -> \
            Dict[int, Union[Dict[Any, Any], Dict[str, InputData], Dict[str, OutputData]]]:

        puts_data = {1: {}, 2: {}}
        for idx, col_name in enumerate(data.inputs, 1):
            parameters_to_pass = {"csv_path": data.csv_path,
                                  "column": col_name,
                                  "type": self.input_type,
                                  "parameters": getattr(val_inputs, f"{self.input_type.value}Validator")(
                                      **data.dict())}
            put_data = InputData(**parameters_to_pass)
            puts_data[1][f"{idx}_{col_name}"] = put_data
        for idx, col_name in enumerate(data.outputs, 1 + len(data.inputs)):
            parameters_to_pass = {"csv_path": data.csv_path,
                                  "column": col_name,
                                  "type": self.output_type,
                                  "parameters": getattr(val_outputs, f"{self.output_type.value}Validator")(
                                      **data.dict())}
            put_data = OutputData(**parameters_to_pass)
            puts_data[2][f"{idx}_{col_name}"] = put_data

        return puts_data


class TimeseriesTrend(Dataframe):
    input_type = LayerInputTypeChoice.Timeseries
    output_type = LayerOutputTypeChoice.Trend

    def __init__(
            self,
            csv_path: str,
            train_size: float,
            inputs: list,
            output: str,
            preprocessing: str,
            length: int,
            step: int,
            deviation: float,
            one_hot_encoding: bool
    ):
        """
        Класс подготовки датасета для задачи временных рядов с предсказанием тренда.

        :param csv_path: Путь до csv-файла;
        :param train_size: Соотношение обучающей выборки к валидационной;
        :param inputs: Список колонок, используемых в качестве входа;
        :param outputs: Колонка, используемая в качестве выхода;
        :param preprocessing: Выбор скейлера. Варианты: "StandardScaler", "MinMaxScaler";
        :param length: Длина одного примера;
        :param step: Шаг движения по аудиофайлу при составлении примеров;
        :param deviation: Отклонение нулевого тренда в процентах;
        :param one_hot_encoding: Перевод Y массивов в формат One-Hot Encoding.
        """
        super().__init__(csv_path=csv_path, inputs=inputs, output=output,  train_size=train_size,
                         preprocessing=preprocessing, length=length, step=step, deviation=deviation,
                         one_hot_encoding=one_hot_encoding)

    def preprocess_put_data(self, data, data_type: LayerSelectTypeChoice) -> \
            Dict[int, Union[Dict[Any, Any], Dict[str, InputData], Dict[str, OutputData]]]:

        puts_data = {1: {}, 2: {}}
        for idx, col_name in enumerate(data.inputs, 1):
            parameters_to_pass = {"csv_path": data.csv_path,
                                  "column": col_name,
                                  "type": self.input_type,
                                  "parameters": getattr(val_inputs, f"{self.input_type.value}Validator")(
                                      **data.dict())}
            put_data = InputData(**parameters_to_pass)
            puts_data[1][f"{idx}_{col_name}"] = put_data
        parameters_to_pass = {"csv_path": data.csv_path,
                              "column": data.output,
                              "type": self.output_type,
                              "parameters": getattr(val_outputs, f"{self.output_type.value}Validator")(
                                  **data.dict())}
        put_data = OutputData(**parameters_to_pass)
        puts_data[2][f"{len(puts_data[1]) + 1}_{data.output}"] = put_data

        return puts_data



from typing import Any

import numpy as np
from gensim.models import Word2Vec
from tensorflow.keras.preprocessing.text import Tokenizer
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from terra_ai_datasets.creation.validators.inputs import TextValidator, ImageValidator


class TerraImageScaler:
    channels = ['red', 'green', 'blue']
    range = (0, 1)

    def __init__(self, height: int, width: int):

        self.shape: tuple = (height, width)
        self.trained_values: dict = {ch: {'min': np.full(self.shape, 255, dtype='uint8'),
                                          'max': np.zeros(self.shape, dtype='uint8')} for ch in self.channels}

    def fit(self, img):
        for i, channel in enumerate(self.channels):
            min_mask = img[:, :, i] < self.trained_values[channel]['min']
            max_mask = img[:, :, i] > self.trained_values[channel]['max']
            self.trained_values[channel]['min'][min_mask] = img[:, :, i][min_mask]
            self.trained_values[channel]['max'][max_mask] = img[:, :, i][max_mask]

    def transform(self, img):

        transformed_img = []
        for ch in self.channels:
            x = img[:, :, self.channels.index(ch)]
            y1 = np.full(self.shape, self.range[0])
            y2 = np.full(self.shape, self.range[1])
            x1 = self.trained_values[ch]['min']
            x2 = self.trained_values[ch]['max']
            y = y1 + ((x - x1) / (x2 - x1)) * (y2 - y1)
            transformed_img.append(y)

        array = np.moveaxis(np.array(transformed_img), 0, -1)

        array[array < self.range[0]] = self.range[0]
        array[array > self.range[1]] = self.range[1]

        if np.isnan(array).any():
            array = np.nan_to_num(array)

        return array

    def inverse_transform(self, img):

        transformed_img = []
        for ch in self.channels:
            x = img[:, :, self.channels.index(ch)]
            x1 = np.full(self.shape, self.range[0])
            x2 = np.full(self.shape, self.range[1])
            y1 = self.trained_values[ch]['min']
            y2 = self.trained_values[ch]['max']
            y = y1 + ((x - x1) / (x2 - x1)) * (y2 - y1)
            transformed_img.append(y)

        array = np.moveaxis(np.array(transformed_img), 0, -1)

        array[array < 0] = 0
        array[array > 255] = 255

        return array.astype(np.uint8)


def create_min_max_scaler(parameters: Any):

    scaler = MinMaxScaler()

    return scaler


def create_standard_scaler(parameters: Any):

    scaler = StandardScaler()

    return scaler


def create_terra_image_scaler(parameters: ImageValidator):

    scaler = TerraImageScaler(height=parameters.height, width=parameters.width)

    return scaler


def create_tokenizer(parameters: TextValidator):
    tokenizer = Tokenizer(**{'num_words': parameters.max_words_count,
                             'filters': parameters.filters,
                             'lower': True,
                             'split': ' ',
                             'char_level': False,
                             'oov_token': '<UNK>'}
                          )
    return tokenizer


def create_bag_of_words(parameters: TextValidator):
    return create_tokenizer(parameters)


def create_embedding(parameters: TextValidator):
    return create_tokenizer(parameters)


def create_word_to_vec(text_list: list, parameters: TextValidator):

    text_list = [elem.split(' ') for elem in text_list]
    word2vec = Word2Vec(text_list, vector_size=parameters.word2vec_size)
    return word2vec


# def inverse_data(self, options: dict):
#     out_dict = {}
#     for put_id, value in options.items():
#         out_dict[put_id] = {}
#         for col_name, array in value.items():
#             if type(self.preprocessing[put_id][col_name]) == StandardScaler or \
#                     type(self.preprocessing[put_id][col_name]) == MinMaxScaler:
#                 out_dict[put_id].update({col_name: self.preprocessing[put_id][col_name].inverse_transform(array)})
#
#             elif type(self.preprocessing[put_id][col_name]) == Tokenizer:
#                 inv_tokenizer = {index: word for word, index in
#                                  self.preprocessing[put_id][col_name].word_index.items()}
#                 out_dict[put_id].update({col_name: ' '.join([inv_tokenizer[seq] for seq in array])})
#
#             else:
#                 out_dict[put_id].update({col_name: ' '.join(
#                     [self.preprocessing[put_id][col_name].most_similar(
#                         positive=[seq], topn=1)[0][0] for seq in array])})
#     return out_dict



from enum import Enum
from typing import Optional, List, Tuple, Dict, Any

from pydantic import BaseModel, validator
from pydantic.types import PositiveInt, PositiveFloat

from terra_ai_datasets.creation.validators import augmentation_params
from terra_ai_datasets.creation.validators.augmentation_params import AugmentationMode


class BaseAugmentation(BaseModel):
    augmentation_coef: PositiveFloat = 0.0

    # @validator("augmentation_coef")
    # def apply_coef_range(cls, v):
    #     if not 0.0 < v:
    #         raise ValueError('Ensure "augmentation_coef" is positive')
    #     return v

    class Config:
        use_enum_values = True


class AugmentationTypes(str, Enum):
    image = "Image"


# --- Image validators ---
class ImageScalers(str, Enum):
    none = None
    min_max_scaler = "MinMaxScaler"
    terra_image_scaler = "TerraImageScaler"


class ImageNetworkTypes(str, Enum):
    linear = "Linear"
    convolutional = "Convolutional"


class ImageProcessTypes(str, Enum):
    stretch = "Stretch"
    fit = "Fit"
    cut = "Cut"


class AugmentationData(BaseAugmentation):
    aug_type: AugmentationTypes
    mode: AugmentationMode
    parameters: Any = None

    @validator("parameters", always=True)
    def validate_date(cls, value, values):
        if values.get("mode"):
            return getattr(augmentation_params, f"{values.get('aug_type')}AugmentationParameters")(
                **getattr(augmentation_params, f"{values.get('aug_type').lower()}_mode_config")[values["mode"]]
            )


class ImageValidator(BaseModel):
    height: PositiveInt
    width: PositiveInt
    network: ImageNetworkTypes = ImageNetworkTypes.convolutional
    process: ImageProcessTypes = ImageProcessTypes.stretch
    preprocessing: ImageScalers = ImageScalers.none
    augmentation: Optional[AugmentationMode]


# --- Text validators ---
class TextModeTypes(str, Enum):
    full = "Full"
    length_and_step = "Length and step"


class TextProcessTypes(str, Enum):
    none = "None"
    embedding = "Embedding"
    bag_of_words = "Bag of words"
    word_to_vec = "Word2Vec"


class TextValidator(BaseModel):
    filters: str = '–—!"#$%&()*+,-./:;<=>?@[\\]^«»№_`{|}~\t\n\xa0–\ufeff'
    preprocessing: TextProcessTypes
    max_words_count: Optional[PositiveInt]
    word2vec_size: Optional[PositiveInt]
    mode: TextModeTypes
    pymorphy: bool
    max_words: Optional[PositiveInt]
    length: Optional[PositiveInt]
    step: Optional[PositiveInt]

    @validator("mode")
    def _validate_mode(cls, value):
        if value == TextModeTypes.full:
            cls.__fields__["max_words"].required = True
            cls.__fields__["length"].required = False
            cls.__fields__["step"].required = False
        elif value == TextModeTypes.length_and_step:
            cls.__fields__["max_words"].required = False
            cls.__fields__["length"].required = True
            cls.__fields__["step"].required = True
        return value

    @validator("preprocessing")
    def _validate_preprocessing(cls, value):
        if value == TextProcessTypes.word_to_vec:
            cls.__fields__["max_words_count"].required = False
            cls.__fields__["word2vec_size"].required = True
        else:
            cls.__fields__["max_words_count"].required = True
            cls.__fields__["word2vec_size"].required = False
        return value


# --- Audio validators ---
class AudioScalers(str, Enum):
    none = None
    min_max_scaler = "MinMaxScaler"
    standard_scaler = "StandardScaler"


class AudioParameterTypes(str, Enum):
    audio_signal = "Audio signal"
    chroma_stft = "Chroma stft"
    mfcc = "MFCC"
    rms = "RMS"
    spectral_centroid = "Spectral centroid"
    spectral_bandwidth = "Spectral bandwidth"
    spectral_rolloff = "Spectral rolloff"
    zero_crossing_rate = "Zero crossing rate"


class AudioResampleTypes(str, Enum):
    kaiser_best = "Kaiser best"
    kaiser_fast = "Kaiser fast"
    scipy = "Scipy"


class AudioFillTypes(str, Enum):
    last_millisecond = "Last millisecond"
    loop = "Loop"


class AudioModeTypes(str, Enum):
    full = "Full"
    length_and_step = "Length and step"


class AudioValidator(BaseModel):
    sample_rate: PositiveInt
    mode: AudioModeTypes
    parameter: List[AudioParameterTypes]
    fill_mode: AudioFillTypes
    resample: AudioResampleTypes
    preprocessing: AudioScalers
    max_seconds: Optional[PositiveFloat]
    length: Optional[PositiveFloat]
    step: Optional[PositiveFloat]

    @validator("mode")
    def _validate_mode(cls, value: AudioModeTypes) -> AudioModeTypes:
        if value == AudioModeTypes.full:
            cls.__fields__["max_seconds"].required = True
            cls.__fields__["length"].required = False
            cls.__fields__["step"].required = False
        elif value == AudioModeTypes.length_and_step:
            cls.__fields__["max_seconds"].required = False
            cls.__fields__["length"].required = True
            cls.__fields__["step"].required = True
        return value


# --- Dataframe validators ---
class CategoricalValidator(BaseModel):
    one_hot_encoding: bool
    classes_names: Optional[List[str]]


class RawValidator(BaseModel):
    pass


# --- Timeseries validators ---
class TimeseriesScalers(str, Enum):
    none = None
    min_max_scaler = "MinMaxScaler"
    standard_scaler = "StandardScaler"


class TimeseriesValidator(BaseModel):
    length: PositiveInt
    step: PositiveInt
    preprocessing: TimeseriesScalers



import numpy as np

from terra_ai_datasets.create import DataframeRegression


def test_regression():

    dataset = DataframeRegression(
        csv_path='./tests/dataset/flats/flats_cut.csv',
        train_size=0.7,
        inputs=[
            {"columns": ["Станция метро", "Пешком или на транспорте"], "type": "Categorical", "parameters": {"one_hot_encoding": True}},
            {"columns": ["Этаж"], "type": "Raw", "parameters": {}},
            {"columns": ["Примечание"], "type": "Text", "parameters": {"max_words_count": 20000, "mode": "Full", "preprocessing": "Embedding", "max_words": 100, "pymorphy": False}},
        ],
        output="Цена, тыс.руб.",
        preprocessing="None"
    )
    dataset.create(use_generator=True)

    for inp, out in dataset.dataset['train'].batch(4):
        assert inp["input_1"].shape == (4, 11), 'Input shape does not match'
        assert inp["input_2"].shape == (4, ), 'Input shape does not match'
        assert inp["input_3"].shape == (4, 100), 'Input shape does not match'
        assert out["output_1"].shape == (4, ), 'Output shape does not match'

        assert inp["input_1"].dtype == np.uint8
        assert inp["input_2"].dtype == np.int32
        assert inp["input_3"].dtype == np.int32
        assert out["output_1"].dtype == np.float64

        break



from enum import Enum
from typing import Dict, Optional, List

from pydantic import BaseModel
from pydantic.types import PositiveInt
from pydantic.color import Color

from terra_ai_datasets.creation.validators.inputs import ImageProcessTypes


class RegressionScalers(str, Enum):
    none = None
    min_max_scaler = "MinMaxScaler"
    standard_scaler = "StandardScaler"


class RegressionValidator(BaseModel):
    preprocessing: RegressionScalers = RegressionScalers.none


class ClassificationValidator(BaseModel):
    one_hot_encoding: bool
    classes_names: Optional[List[str]]


class SegmentationValidator(BaseModel):
    rgb_range: PositiveInt
    classes: Dict[str, Color]
    height: Optional[PositiveInt]
    width: Optional[PositiveInt]
    process: Optional[ImageProcessTypes]


# --- Timeseries validators ---
class DepthScalers(str, Enum):
    none = None
    min_max_scaler = "MinMaxScaler"
    standard_scaler = "StandardScaler"


class DepthValidator(BaseModel):
    depth: PositiveInt
    preprocessing: DepthScalers
    length: Optional[PositiveInt]
    step: Optional[PositiveInt]


class TrendValidator(BaseModel):
    deviation: float
    one_hot_encoding: bool
    classes_names: List[str] = ["Не изменился", "Вверх", "Вниз"]
    length: Optional[PositiveInt]
    step: Optional[PositiveInt]



from pathlib import Path
import json

import pandas as pd
import joblib

from terra_ai_datasets.creation.dataset import TerraDataset
from terra_ai_datasets.creation.utils import DatasetPathsData
from terra_ai_datasets.creation.validators import creation_data, inputs, outputs
from terra_ai_datasets.creation.validators.structure import DatasetData
from terra_ai_datasets.creation.validators.tasks import LayerInputTypeChoice


class LoadDataset(TerraDataset):

    def __init__(self, path_to_dataset: str):
        super().__init__()
        path_to_dataset = Path(path_to_dataset)
        assert path_to_dataset.is_dir(), "Датасет по указанному пути не найден"

        dataset_paths_data = DatasetPathsData(Path(path_to_dataset))

        for split in dataset_paths_data.instructions.dataframe.iterdir():
            self.dataframe[split.stem] = pd.read_csv(split, index_col=0)

        for instr_path in dataset_paths_data.instructions.parameters.iterdir():
            put, put_id, put_type = instr_path.stem.split('_')
            if not self.put_instructions.get(int(put_id)):
                self.put_instructions[int(put_id)] = {}
            with open(instr_path, 'r') as conf:
                dict_data = json.loads(json.load(conf))
                self.put_instructions[int(put_id)][f"{put_id}_{put_type}"] = \
                    getattr(creation_data, f"{put.capitalize()}InstructionsData")(**dict_data)
                put = self.put_instructions[int(put_id)][f'{put_id}_{put_type}']
                self.put_instructions[int(put_id)][f"{put_id}_{put_type}"].parameters = \
                    getattr(inputs, f"{put.type}Validator")(**dict_data['parameters']) \
                    if put.type in LayerInputTypeChoice\
                    else getattr(outputs, f"{put.type}Validator")(**dict_data['parameters'])

        with open(dataset_paths_data.config, 'r') as conf:
            self.dataset_data = DatasetData(**json.loads(json.load(conf)))

        for prep_path in dataset_paths_data.preprocessing.iterdir():
            self.preprocessing[prep_path.stem] = joblib.load(prep_path)

        if self.dataset_data.is_created:
            for split, dataframe in self.dataframe.items():
                if self.dataset_data.use_generator:
                    self._dataset[split] = self.create_dataset_object_from_instructions(
                        self.put_instructions, dataframe
                    )
                else:
                    for array_path in dataset_paths_data.arrays.inputs.iterdir():
                        put_id, put = array_path.stem.split('_')
                        self.X[put][int(put_id)] = joblib.load(array_path)
                    for array_path in dataset_paths_data.arrays.outputs.iterdir():
                        put_id, put = array_path.stem.split('_')
                        self.Y[put][int(put_id)] = joblib.load(array_path)
                    self._dataset[split] = self.create_dataset_object_from_arrays(self.X[split], self.Y[split])



import numpy as np

from terra_ai_datasets.create import ImageSegmentation


def test_segmentation_minmax_automatic():

    dataset = ImageSegmentation(source_path=['./airplane/Самолеты'], target_path=['./airplane/Сегменты'], train_size=.8,
                                width=64, height=80, preprocessing="MinMaxScaler", network="Convolutional",
                                process='Fit', rgb_range=50, num_classes=2)

    dataset.create()

    for inp, out in dataset.dataset['train'].batch(16):
        assert inp["input_1"].shape == (16, 80, 64, 3), 'Input shape does not match'
        assert out["output_1"].shape == (16, 80, 64, 2), 'Output shape does not match'

        assert inp["input_1"].dtype == np.float32
        assert out["output_1"].dtype == np.uint8

        break


def test_segmentation_no_scaler_from_file():
    dataset = ImageSegmentation(source_path=['./airplane/Самолеты'], target_path=['./airplane/Сегменты'], train_size=.8,
                                width=32, height=40, preprocessing="None", network="Convolutional",
                                process='Fit', rgb_range=50, classes_path='./airplane/labelmap.txt')

    dataset.create(use_generator=True)

    for inp, out in dataset.dataset['train'].batch(16):
        assert inp["input_1"].shape == (16, 40, 32, 3), 'Input shape does not match'
        assert out["output_1"].shape == (16, 40, 32, 2), 'Output shape does not match'

        assert inp["input_1"].dtype == np.uint8
        assert out["output_1"].dtype == np.uint8

        break



import random

from IPython.display import Audio, display
import matplotlib.pyplot as plt
import numpy as np
from imgaug import SegmentationMapsOnImage

from terra_ai_datasets.creation.augmentation import create_image_augmentation
from terra_ai_datasets.creation.utils import create_put_array, preprocess_put_array


def visualize_image_classification(dataframe, put_instructions, preprocessing, augmentation):

    x_data = dataframe.loc[:, "1_Image"].tolist()
    y_data = dataframe.loc[:, "2_Classification"].tolist()

    x_instructions = put_instructions[1]['1_Image']
    y_instructions = put_instructions[2]['2_Classification']

    classes_names = y_instructions.parameters.classes_names

    img_aug = augmentation.get("1_Image")

    fig, ax = plt.subplots(1, len(classes_names), figsize=(3 * len(classes_names), 5))
    for i, cls_name in enumerate(classes_names):
        sample_idx = random.choice([idx for idx, cl_n in enumerate(y_data) if cl_n == cls_name])
        x_sample = x_data[sample_idx]
        y_sample = y_data[sample_idx]

        x_array = create_put_array(x_sample, x_instructions)
        y_array = create_put_array(y_sample, y_instructions)

        if img_aug and "augm" in x_sample:
            x_array = img_aug(image=x_array)

        if preprocessing.get("1_Image"):
            x_array = preprocess_put_array(
                x_array, x_instructions, preprocessing["1_Image"]
            )

        ax[i].imshow(x_array)
        ax[i].set_title(f"{cls_name} - {str(y_array)}")

    plt.show()


def visualize_image_segmentation(dataframe, put_instructions, preprocessing, augmentation):

    sample_idx = random.randint(0, len(dataframe) - 1)

    img_aug = augmentation.get("1_Image")

    x_sample = str(dataframe.loc[sample_idx, "1_Image"])
    y_sample = str(dataframe.loc[sample_idx, "2_Segmentation"])

    x_instructions = put_instructions[1]['1_Image']
    y_instructions = put_instructions[2]['2_Segmentation']

    x_array = create_put_array(x_sample, x_instructions)
    y_array = create_put_array(y_sample, y_instructions)

    if img_aug and "augm" in x_sample:
        y_maps = SegmentationMapsOnImage(y_array, shape=y_array.shape)
        x_array, y_array = img_aug(image=x_array, segmentation_maps=y_maps)
        y_array = y_array.arr

    if preprocessing.get("1_Image"):
        x_array = preprocess_put_array(
            x_array, x_instructions, preprocessing["1_Image"]
        )

    num_classes = len(y_instructions.parameters.classes.items())
    fig, ax = plt.subplots(1, num_classes+1, figsize=(5*num_classes+1, 5))
    ax[0].imshow(x_array)
    ax[0].set_title(f"Вход")

    for i, (class_name, color) in enumerate(y_instructions.parameters.classes.items()):
        ax[i + 1].imshow(y_array[:, :, i])
        ax[i + 1].set_title(f"{class_name} - {str(color.as_rgb())}")

    plt.show()


def visualize_timeseries_depth(dataframe, put_instructions, preprocessing, **kwargs):

    plt.figure(figsize=(15, 5))
    sample_idx = None
    for col_name, col_data in put_instructions[1].items():
        x_instructions = put_instructions[1][col_name]
        if sample_idx is None:
            sample_idx = random.randint(0, len(dataframe) // x_instructions.parameters.length - 10)
        x_data = dataframe.loc[:, col_name][sample_idx*x_instructions.parameters.length: sample_idx*x_instructions.parameters.length + x_instructions.parameters.length + 1].tolist()
        x_array = create_put_array(x_data, x_instructions)
        if preprocessing.get(col_name):
            x_array = preprocess_put_array(
                x_array, x_instructions, preprocessing[col_name]
            )
        plt.plot(x_array.flatten(), label=f"{col_name} (inp)")

    for col_name, col_data in put_instructions[2].items():
        y_instructions = put_instructions[2][col_name]
        y_data = dataframe.loc[:, col_name][sample_idx*y_instructions.parameters.length + y_instructions.parameters.length: sample_idx*y_instructions.parameters.length + y_instructions.parameters.length + y_instructions.parameters.depth].tolist()
        y_array = create_put_array(y_data, y_instructions)
        if preprocessing.get(col_name):
            y_array = preprocess_put_array(
                y_array, y_instructions, preprocessing[col_name]
            )
        plt.plot([None for _ in range(col_data.parameters.length)] + y_array[:, 0].tolist(), label=f"{col_name} (out)")
        plt.axvline(x=y_instructions.parameters.length, color='r', linestyle="--", linewidth=2)

    plt.grid(True, alpha=0.3)
    plt.legend()


def visualize_timeseries_trend(dataframe, put_instructions, preprocessing, **kwargs):

    plt.figure(figsize=(15, 5))
    sample_idx = None
    for col_name, col_data in put_instructions[1].items():
        x_instructions = put_instructions[1][col_name]

        if sample_idx is None:
            sample_idx = random.randint(0, len(dataframe) // x_instructions.parameters.length - 10)
        x_data = dataframe.loc[:, col_name][sample_idx*x_instructions.parameters.length: sample_idx*x_instructions.parameters.length + x_instructions.parameters.length].tolist()
        x_array = create_put_array(x_data, x_instructions)
        if preprocessing.get(col_name):
            x_array = preprocess_put_array(
                x_array, x_instructions, preprocessing[col_name]
            )
        plt.plot(x_array.flatten(), label=f"{col_name} (inp)")

    for col_name, col_data in put_instructions[2].items():
        y_instructions = put_instructions[2][col_name]
        y_data = dataframe.loc[:, col_name][sample_idx*y_instructions.parameters.length + y_instructions.parameters.length - 1: sample_idx*y_instructions.parameters.length + y_instructions.parameters.length + 1].tolist()
        y_array = create_put_array(y_data, y_instructions)
        if y_instructions.parameters.one_hot_encoding:
            y_array = np.argmax(y_array)
        if y_array == 1:
            color = "green"
        elif y_array == 2:
            color = "red"
        else:
            color = "cyan"
        start_idx, stop_idx = y_data
        plt.plot([None for _ in range(y_instructions.parameters.length)] + [stop_idx], marker="o", markersize=2,
                 markeredgecolor=color, markerfacecolor=color)
        plt.arrow(y_instructions.parameters.length - 1, start_idx, 1, stop_idx - start_idx, width=0.2, color=color)
        plt.grid(alpha=0.3)
        plt.legend()
        plt.show()


def visualize_text_classification(dataframe, put_instructions, **kwargs):

    x_data = dataframe.loc[:, "1_Text"].tolist()
    y_data = dataframe.loc[:, "2_Classification"].tolist()

    classes_names = put_instructions[2]['2_Classification'].parameters.classes_names

    for i, cls_name in enumerate(classes_names):
        sample_idx = random.choice([idx for idx, cl_n in enumerate(y_data) if cl_n == cls_name])
        x_sample = x_data[sample_idx]
        y_sample = y_data[sample_idx]

        print(f"Класс: \033[1m{y_sample}\033[0m")
        display(x_sample)
        if i + 1 != len(classes_names):
            print()


def visualize_audio_classification(dataframe, put_instructions, **kwargs):

    cls_idx = len(dataframe.columns.tolist())
    x_data = dataframe.loc[:, "1_Audio"].tolist()
    y_data = dataframe.loc[:, f"{cls_idx}_Classification"].tolist()

    x_instructions = put_instructions[1]['1_Audio']
    y_instructions = put_instructions[cls_idx][f'{cls_idx}_Classification']

    sample_rate = x_instructions.parameters.sample_rate
    classes_names = y_instructions.parameters.classes_names

    for i, cls_name in enumerate(classes_names):
        sample_idx = random.choice([idx for idx, cl_n in enumerate(y_data) if cl_n == cls_name])
        x_sample = x_data[sample_idx]
        y_sample = y_data[sample_idx]

        x_array = create_put_array(x_sample, x_instructions)

        file_path, rng = x_sample.split(';')

        print(f"Класс: \033[1m{y_sample}\033[0m\nФайл: {file_path}\nОтрезок: {rng}")
        display(Audio(data=x_array, rate=sample_rate))
        if i + 1 != len(classes_names):
            print()


def visualize_dataframe_classification(dataframe, put_instructions, preprocessing):
    pass


def visualize_dataframe_regression(dataframe, put_instructions, preprocessing):
    pass



from enum import Enum


class TasksChoice(str, Enum):
    ImageClassification = "ImageClassification"
    ImageSegmentation = "ImageSegmentation"
    TextClassification = "TextClassification"
    AudioClassification = "AudioClassification"
    DataframeClassification = "DataframeClassification"
    DataframeRegression = "DataframeRegression"
    TimeseriesDepth = "TimeseriesDepth"
    TimeseriesTrend = "TimeseriesTrend"


class LayerInputTypeChoice(str, Enum):
    Image = "Image"
    Text = "Text"
    Audio = "Audio"
    Dataframe = "Dataframe"
    Categorical = "Categorical"
    Raw = "Raw"
    Timeseries = "Timeseries"
    # Scaler = "Scaler"


class LayerOutputTypeChoice(str, Enum):
    Classification = "Classification"
    Segmentation = "Segmentation"
    Regression = "Regression"
    Dataset = "Dataset"
    Trend = "Trend"
    Depth = "Depth"


class LayerSelectTypeChoice(str, Enum):
    table = "table"
    folder = "folder"



import numpy as np

from terra_ai_datasets.create import TimeseriesDepth, TimeseriesTrend


def test_timeseries_depth():

    dataset = TimeseriesDepth(
        csv_path='./tests/dataset/shares/YNDX_1d.csv',
        train_size=0.7,
        inputs=["<OPEN>", "<HIGH>", "<LOW>", "<CLOSE>"],
        outputs=["<LOW>", "<CLOSE>"],
        preprocessing="None",
        length=30,
        step=7,
        depth=10,
    )
    dataset.create(use_generator=True)

    for inp, out in dataset.dataset['train'].batch(4):
        assert inp["input_1"].shape == (4, 30, 4), 'Input shape does not match'
        assert out["output_1"].shape == (4, 10, 2), 'Output shape does not match'

        assert inp["input_1"].dtype == np.float64
        assert out["output_1"].dtype == np.float64

        break


def test_timeseries_trend():

    dataset = TimeseriesTrend(
        csv_path='shares/YNDX_1d.csv',
        train_size=0.7,
        inputs=["<OPEN>", "<HIGH>", "<LOW>", "<CLOSE>"],
        output="<LOW>",
        preprocessing="None",
        length=100,
        step=30,
        deviation=.5,
        one_hot_encoding=False
    )
    dataset.create()

    for inp, out in dataset.dataset['train'].batch(4):
        assert inp["input_1"].shape == (4, 100, 4), 'Input shape does not match'
        assert out["output_1"].shape == (4, ), 'Output shape does not match'

        assert inp["input_1"].dtype == np.float64
        assert out["output_1"].dtype == np.int32

        break



import numpy as np

from terra_ai_datasets.create import ImageClassification, TextClassification, AudioClassification, \
    DataframeClassification


def test_image_classification_minmax():

    dataset = ImageClassification(source_path=['./tests/dataset/cars/Мерседес', './tests/dataset/cars/Феррари'],
                                  width=80, height=64, preprocessing='MinMaxScaler', network="Linear", process='Fit',
                                  train_size=0.7, one_hot_encoding=True)

    dataset.create(use_generator=True)

    for inp, out in dataset.dataset['train'].batch(4):
        assert inp["input_1"].shape == (4, 64*80*3), 'Input shape does not match'
        assert out["output_1"].shape == (4, 2), 'Output shape does not match'

        assert inp["input_1"].dtype == np.float32, 'Input outputs wrong datatype'
        assert out["output_1"].dtype == np.uint8, 'Output outputs wrong datatype'

        break


def test_image_classification_terra_image():
    dataset = ImageClassification(source_path=['./tests/dataset/cars'], width=160, height=128,
                                  preprocessing='TerraImageScaler', network="Convolutional", process='Stretch',
                                  train_size=0.7, one_hot_encoding=False)

    dataset.create()

    for inp, out in dataset.dataset['train'].batch(6):
        assert inp["input_1"].shape == (6, 128, 160, 3), 'Input shape does not match'
        assert out["output_1"].shape == (6, ), 'Output shape does not match'

        assert inp["input_1"].dtype == np.float32
        assert out["output_1"].dtype == np.uint8

        break


def test_image_classification_no_scaler():
    dataset = ImageClassification(source_path=['./tests/dataset/cars/Мерседес', './tests/dataset/cars/Рено',
                                               './tests/dataset/cars/Феррари'], width=80, height=64,
                                  preprocessing='None', network="Convolutional", process='Cut',
                                  train_size=0.7, one_hot_encoding=True)

    dataset.create()

    for inp, out in dataset.dataset['train'].batch(6):
        assert inp["input_1"].shape == (6, 64, 80, 3), 'Input shape does not match'
        assert out["output_1"].shape == (6, 3), 'Output shape does not match'

        assert inp["input_1"].dtype == np.uint8
        assert out["output_1"].dtype == np.uint8

        break


def test_text_classification_emb():

    dataset = TextClassification(
        source_path=['./tests/dataset/symptoms/Аппендицит', './tests/dataset/symptoms/Гастрит',
                     './tests/dataset/symptoms/Гепатит'],
        train_size=0.7,
        max_words_count=20000,
        mode="Length and step",
        preprocessing="Embedding",
        length=100,
        step=30,
        one_hot_encoding=True,
        pymorphy=True
    )
    dataset.create()

    for inp, out in dataset.dataset['train'].batch(2):
        assert inp["input_1"].shape == (2, 100), 'Input shape does not match'
        assert out["output_1"].shape == (2, 3), 'Output shape does not match'

        assert inp["input_1"].dtype == np.int32
        assert out["output_1"].dtype == np.uint8

        break


def test_text_classification_bow():

    dataset = TextClassification(
        source_path=['./symptoms'],
        train_size=0.7,
        max_words_count=10000,
        mode="Full",
        preprocessing="Bag of words",
        max_words=2000,
        one_hot_encoding=True,
        pymorphy=False
    )

    dataset.create(use_generator=True)

    for inp, out in dataset.dataset['train'].batch(3):
        assert inp["input_1"].shape == (3, 10000), 'Input shape does not match'
        assert out["output_1"].shape == (3, 10), 'Output shape does not match'

        assert inp["input_1"].dtype == np.int8
        assert out["output_1"].dtype == np.uint8

        break


def test_text_classification_w2v():

    dataset = TextClassification(
        source_path=['./symptoms'],
        train_size=0.7,
        max_words_count=10000,
        word2vec_size=50,
        mode="Length and step",
        preprocessing="Word2Vec",
        length=100,
        step=30,
        one_hot_encoding=True,
        pymorphy=False
    )

    dataset.create(use_generator=True)

    for inp, out in dataset.dataset['train'].batch(16):
        assert inp["input_1"].shape == (16, 100, 50), 'Input shape does not match'
        assert out["output_1"].shape == (16, 10), 'Output shape does not match'

        assert inp["input_1"].dtype == np.float64
        assert out["output_1"].dtype == np.uint8

        break


def test_audio_classification():

    dataset = AudioClassification(
        source_path=['./smarthome/1_Кондиционер', './smarthome/2_Свет'],
        train_size=0.7,
        sample_rate=22050,
        mode="Length and step",
        parameter=["Audio signal", "MFCC", "RMS"],
        fill_mode="Last millisecond",
        resample="Kaiser fast",
        preprocessing="None",
        max_seconds=2,
        length=0.5,
        step=0.4,
        one_hot_encoding=True
        )

    dataset.create(use_generator=True)

    for inp, out in dataset.dataset['train'].batch(16):
        assert inp["input_1"].shape == (16, 11025), 'Input shape does not match'
        assert inp["input_2"].shape == (16, 22, 20), 'Input shape does not match'
        assert inp["input_3"].shape == (16, 22), 'Input shape does not match'
        assert out["output_1"].shape == (16, 2), 'Output shape does not match'

        assert inp["input_1"].dtype == np.float32
        assert inp["input_2"].dtype == np.float32
        assert inp["input_3"].dtype == np.float32
        assert out["output_1"].dtype == np.uint8

        break


def test_dataframe_classification():
    dataset = DataframeClassification(
        csv_path='./tests/dataset/flats/flats_cut.csv',
        train_size=0.5,
        inputs=[
            {"columns": ["Станция метро", "Пешком или на транспорте"], "type": "Categorical", "parameters": {"one_hot_encoding": True}},
            {"columns": ["Этаж"], "type": "Raw", "parameters": {}},
            {"columns": ["Примечание"], "type": "Text", "parameters": {"max_words_count": 20, "mode": "Full", "preprocessing": "Embedding", "max_words": 10, "pymorphy": False}},
        ],
        output="Тип дома",
        one_hot_encoding=False
    )
    dataset.create()

    for inp, out in dataset.dataset['train'].batch(4):
        assert inp["input_1"].shape == (4, 11), 'Input shape does not match'
        assert inp["input_2"].shape == (4, ), 'Input shape does not match'
        assert inp["input_3"].shape == (4, 10), 'Input shape does not match'
        assert out["output_1"].shape == (4, ), 'Output shape does not match'

        assert inp["input_1"].dtype == np.uint8
        assert inp["input_2"].dtype == np.int32
        assert inp["input_3"].dtype == np.int32
        assert out["output_1"].dtype == np.uint8

        break



from setuptools import setup, find_packages
from terra_ai_datasets import __version__

DESCRIPTION = "Framework to create a dataset to train a neural network."
LONG_DESCRIPTION = "terra_ai_datasets is a framework to create " \
                   "a dataset to train a neural network model based on a Keras."

setup(
    name="terra_ai_datasets_framework",
    version=__version__,
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    install_requires=[
        "pydantic>=1.8.2",
        "pandas>=1.3.5",
        "opencv-python>=4.6.0.66",
        "librosa>=0.8.1",
        "pillow>=7.1.2",
        "tqdm>=4.64.1",
        "scikit-learn>=1.0.2",
        "tensorflow>=2.0",
        "joblib>=1.1.0",
        "pymorphy2>=0.9.1",
        "gensim>=3.6.0",
        "gensim>=3.6.0",
    ],
)



import json
from pathlib import Path
from datetime import datetime
from typing import Union, Dict, List, Any, NoReturn

from IPython.display import display
from sklearn.utils import resample
from tensorflow.python.data.ops.dataset_ops import DatasetV2 as Dataset
from tensorflow import TensorSpec
from tqdm import tqdm
import numpy as np
import pandas as pd
import joblib

from terra_ai_datasets.creation import preprocessings, utils, visualize, augmentation
from terra_ai_datasets.creation.augmentation import create_image_augmentation
from terra_ai_datasets.creation.utils import apply_pymorphy, process_directory_paths, decamelize, create_put_array, \
    preprocess_put_array
from terra_ai_datasets.creation.validators import creation_data
from terra_ai_datasets.creation.validators.creation_data import InputData, OutputData, InputInstructionsData, \
    OutputInstructionsData
from terra_ai_datasets.creation.validators import dataset
from terra_ai_datasets.creation.validators import inputs, outputs
from terra_ai_datasets.creation.validators.inputs import CategoricalValidator, ImageScalers, TextProcessTypes, \
    AugmentationData
from terra_ai_datasets.creation.validators.structure import DatasetData

from terra_ai_datasets.creation.validators.tasks import LayerInputTypeChoice, LayerOutputTypeChoice, \
    LayerSelectTypeChoice


def postprocess_put_array(put_arr: list, put_data_type: Union[LayerInputTypeChoice, LayerOutputTypeChoice]) \
        -> np.ndarray:
    if put_data_type in [LayerInputTypeChoice.Timeseries, LayerOutputTypeChoice.Depth]:
        # put_arr = np.moveaxis(np.array(put_arr), [1, 2], [0, 1])
        put_arr = np.concatenate(put_arr, axis=2)
        pass
    # arrays
    elif all([len(np.array(arr).shape) >= 2 for arr in put_arr]):
        put_arr = np.concatenate(put_arr, axis=1)
    # vectors
    elif all([len(np.array(arr).shape) == 1 for arr in put_arr]):
        put_arr = np.concatenate(put_arr, axis=0)
    # scalar
    else:
        put_arr = np.array(put_arr[0])
    return put_arr


class TerraDataset:

    def __init__(self):
        self.X: Dict[str, Dict[str, np.ndarray]] = {"train": {}, "val": {}}
        self.Y: Dict[str, Dict[str, np.ndarray]] = {"train": {}, "val": {}}
        self.preprocessing: Dict[str, Any] = {}
        self.augmentation: Dict[str, Any] = {}
        self.dataframe: Dict[str, pd.DataFrame] = {}
        self.dataset_data: DatasetData = None
        self.put_instructions: Dict[int, Dict[str, Union[InputInstructionsData, OutputInstructionsData]]] = {}

        self._dataset: Dict[str, Dataset] = {"train": None, "val": None}

    @property
    def dataset(self):
        if not self._dataset["train"] and not self._dataset["val"]:
            return 'Массивы не были созданы. Вызовите метод .create(use_generator: bool = False)'
        return self._dataset

    @property
    def train(self):
        if not self._dataset["train"]:
            return 'Массивы не были созданы. Вызовите метод .create(use_generator: bool = False)'
        return self._dataset["train"]

    @property
    def val(self):
        if not self._dataset["val"]:
            return 'Массивы не были созданы. Вызовите метод .create(use_generator: bool = False)'
        return self._dataset["val"]

    @staticmethod
    def create_dataset_object_from_arrays(x_arrays: Dict[str, np.ndarray], y_arrays: Dict[str, np.ndarray]) -> Dataset:
        return Dataset.from_tensor_slices((x_arrays, y_arrays))

    def create_dataset_object_from_instructions(self, put_instr, dataframe) -> Dataset:

        output_signature = [{}, {}]
        length, step, offset, total_samples = 1, 1, 0, len(dataframe)
        inp_id, out_id = 1, 1
        for put_id, cols_dict in put_instr.items():
            put_array = []
            for col_name, put_data in cols_dict.items():
                if put_data.type in \
                        [LayerInputTypeChoice.Timeseries, LayerOutputTypeChoice.Depth, LayerOutputTypeChoice.Trend]:
                    length = put_data.parameters.length
                    total_samples -= length * 2
                    if put_data.type in [LayerOutputTypeChoice.Depth, LayerOutputTypeChoice.Trend]:
                        offset = put_data.parameters.length
                        if put_data.type == LayerOutputTypeChoice.Trend:
                            offset -= 1
                data_to_send = dataframe.loc[0 + offset:0 + offset + length - 1, col_name].to_list()
                data_to_send = data_to_send[0] if len(data_to_send) == 1 else data_to_send
                sample_array = create_put_array(data_to_send, put_data)
                if self.preprocessing.get(col_name):
                    sample_array = preprocess_put_array(
                        sample_array, put_data, self.preprocessing[col_name]
                    )
                put_array.append(
                    sample_array[0] if type(sample_array) == np.ndarray and len(sample_array) == 1 else sample_array
                )
            if put_data.type in [LayerInputTypeChoice.Timeseries, LayerOutputTypeChoice.Depth]:
                put_array = np.expand_dims(np.array(put_array), 1)
                put_array = postprocess_put_array(put_array, put_data.type)[0]
            else:
                put_array = postprocess_put_array(put_array, put_data.type)

            if put_data.type in LayerInputTypeChoice:
                output_signature[0][f"input_{inp_id}"] = TensorSpec(shape=put_array.shape, dtype=put_array.dtype)
                inp_id += 1
            else:
                output_signature[1][f"output_{out_id}"] = TensorSpec(shape=put_array.shape, dtype=put_array.dtype)
                out_id += 1

        return Dataset.from_generator(lambda: self.generator(put_instr, dataframe, length, step, total_samples),
                                      output_signature=tuple(output_signature))

    def generator(self, put_instr, dataframe: pd.DataFrame, length, step, total_samples):

        for i in range(0, total_samples, step):
            inp_id, out_id = 1, 1
            inp_dict, out_dict = {}, {}
            for put_id, cols_dict in put_instr.items():
                put_array = []
                for col_name, put_data in cols_dict.items():
                    offset = 0
                    if put_data.type in [LayerOutputTypeChoice.Depth, LayerOutputTypeChoice.Trend]:
                        offset = put_data.parameters.length
                    data_to_send = dataframe.loc[i+offset: i+offset+length-1, col_name].to_list()
                    data_to_send = data_to_send[0] if len(data_to_send) == 1 else data_to_send
                    sample_array = create_put_array(data_to_send, put_data)

                    # sample_array = create_put_array(dataframe.loc[i, col_name], put_data)
                    if self.preprocessing.get(col_name):
                        sample_array = preprocess_put_array(
                            sample_array, put_data, self.preprocessing[col_name]
                        )
                    put_array.append(
                        sample_array[0] if type(sample_array) == np.ndarray and len(sample_array) == 1 else sample_array
                    )

                if put_data.type in [LayerInputTypeChoice.Timeseries, LayerOutputTypeChoice.Depth]:
                    put_array = np.expand_dims(np.array(put_array), 1)
                    put_array = postprocess_put_array(put_array, put_data.type)[0]
                else:
                    put_array = postprocess_put_array(put_array, put_data.type)

                if put_data.type in LayerInputTypeChoice:
                    inp_dict[f"input_{inp_id}"] = put_array
                    inp_id += 1
                else:
                    out_dict[f"output_{out_id}"] = put_array
                    out_id += 1

            yield inp_dict, out_dict

    def create(self, use_generator: bool = False, verbose: int = 1):

        if verbose not in [0, 1]:
            raise ValueError("Параметр verbose принимает значение 0 или 1")

        if verbose == 0:
            disable = True
            utils.logger.info(f"Выполняется процесс создания датасетов")
        else:
            disable = False

        for split, dataframe in self.dataframe.items():
            inp_id, out_id = 1, 1
            for put_id, cols_dict in self.put_instructions.items():
                if use_generator and split != "train":
                    continue
                for col_name, put_data in cols_dict.items():
                    if split == "train":
                        if "preprocessing" in put_data.parameters.dict() and put_data.parameters.preprocessing.value != 'None':
                            create_preprocessing_parameters = {"parameters": put_data.parameters}
                            if put_data.parameters.preprocessing == TextProcessTypes.word_to_vec:
                                create_preprocessing_parameters.update(
                                    {"text_list": dataframe[col_name].to_list()}
                                )
                            self.preprocessing[col_name] = \
                                getattr(preprocessings, f"create_{put_data.parameters.preprocessing.name}")(
                                    **create_preprocessing_parameters
                                )

                put_array = []
                for col_name, put_data in cols_dict.items():
                    col_array = []
                    length, step, offset, total_samples = 1, 1, 0, len(dataframe)
                    if put_data.type in \
                            [LayerInputTypeChoice.Timeseries, LayerOutputTypeChoice.Depth, LayerOutputTypeChoice.Trend]:
                        length = put_data.parameters.length
                        step = put_data.parameters.step
                        total_samples -= length * 2
                        if put_data.type in [LayerOutputTypeChoice.Depth, LayerOutputTypeChoice.Trend]:
                            offset = put_data.parameters.length
                            if put_data.type == LayerOutputTypeChoice.Trend:
                                offset -= 1
                    for row_idx in tqdm(
                            range(0, total_samples, step),
                            desc=f"{datetime.now().strftime('%H:%M:%S')} | Формирование массивов {split} - {put_data.type} - {col_name}",
                            disable=disable
                    ):
                        data_to_send = dataframe.loc[row_idx+offset:row_idx+offset+length-1, col_name].to_list()
                        data_to_send = data_to_send[0] if len(data_to_send) == 1 else data_to_send
                        sample_array = create_put_array(data_to_send, put_data)
                        if self.preprocessing.get(col_name) and split == "train":
                            if put_data.type == LayerInputTypeChoice.Text:
                                pass
                            elif put_data.type == LayerInputTypeChoice.Image and put_data.parameters.preprocessing == ImageScalers.terra_image_scaler:
                                self.preprocessing[col_name].fit(sample_array)
                            else:
                                self.preprocessing[col_name].partial_fit(sample_array.reshape(-1, 1))
                        if not use_generator or use_generator and put_data.type == LayerInputTypeChoice.Text:
                            col_array.append(
                                sample_array[0] if type(sample_array) == np.ndarray and len(sample_array) == 1 else sample_array
                            )

                    if self.preprocessing.get(col_name) and split == "train":
                        if put_data.type == LayerInputTypeChoice.Text:
                            if put_data.parameters.preprocessing in [TextProcessTypes.embedding, TextProcessTypes.bag_of_words]:
                                self.preprocessing[col_name].fit_on_texts(col_array)

                    if self.augmentation.get(col_name) and split == "train":
                        if put_data.type == LayerInputTypeChoice.Image:
                            aug_idx = dataframe[col_name].str.contains("augm").tolist()
                            if not self.output_type == LayerOutputTypeChoice.Segmentation:
                                for i, val in enumerate(aug_idx):
                                    if val:
                                        col_array[i] = self.augmentation[col_name](image=np.array(col_array[i]))

                    if self.preprocessing.get(col_name) and not use_generator:
                        col_array = preprocess_put_array(
                            np.array(col_array), put_data, self.preprocessing[col_name]
                        )
                    if not use_generator:
                        put_array.append(col_array if len(col_array) > 1 else col_array[0])

                if not use_generator:
                    put_array = postprocess_put_array(put_array, put_data.type)
                    if isinstance(put_data, InputInstructionsData):
                        self.X[split][f"input_{inp_id}"] = put_array
                        inp_id += 1
                    else:
                        self.Y[split][f"output_{out_id}"] = put_array
                        out_id += 1

            if not use_generator:
                self._dataset[split] = self.create_dataset_object_from_arrays(self.X[split], self.Y[split])
            else:
                self._dataset[split] = self.create_dataset_object_from_instructions(
                    self.put_instructions, dataframe
                )
        self.dataset_data.is_created = True
        if verbose == 0:
            utils.logger.info(f"Готово!")

    def visualize(self):
        getattr(visualize, f"visualize_{decamelize(self.dataset_data.task)}")(
            dataframe=self.dataframe["train"],
            put_instructions=self.put_instructions,
            preprocessing=self.preprocessing,
            augmentation=self.augmentation
        )

    def evaluate_on_model(self, model, batch_size: int):
        for inp, out in self.dataset["val"].batch(batch_size):
            pred = model.predict(inp)
            yield inp, out, pred

    def create_input_array(self, *args: List[str]) -> Dict[str, np.ndarray]:
        args_idx = 0
        input_dict = {}
        for put_id, cols_dict in self.put_instructions.items():
            put_array = []
            add_to_input_dict = False
            for col_name, put_data in cols_dict.items():
                if put_data.type in LayerInputTypeChoice:
                    data_to_send = args[args_idx]
                    if put_data.type == LayerInputTypeChoice.Text:
                        data_to_send = [data_to_send]
                    col_array = create_put_array(data_to_send, put_data)
                    if self.augmentation.get(col_name):
                        col_array = self.augmentation[col_name](image=col_array)
                    if self.preprocessing.get(col_name):
                        col_array = preprocess_put_array(
                            np.array(col_array), put_data, self.preprocessing[col_name]
                        )
                    put_array.append(col_array if len(col_array) > 1 else col_array[0])
                    args_idx += 1
                    add_to_input_dict = True
            if add_to_input_dict:
                put_array = postprocess_put_array(put_array, put_data.type)
                input_dict[f"input_{put_id}"] = np.expand_dims(np.array(put_array), 0)
        return input_dict

    def summary(self):

        display(self.dataframe['train'].head())
        print(f"\n\033[1mКол-во примеров в train выборке:\033[0m {len(self.dataframe['train'])}\n"
              f"\033[1mКол-во примеров в val выборке:\033[0m {len(self.dataframe['val'])}")
        print()
        if self.dataset_data.is_created:
            for inp_id, array in enumerate(self.X["train"].values(), 1):
                print(f"\033[1mРазмерность входного массива {inp_id}:\033[0m", array[0].shape)
            for out_id, array in enumerate(self.Y["train"].values(), 1):
                print(f"\033[1mРазмерность выходного массива {out_id}:\033[0m", array[0].shape)

        put_id, col_name, output_type = None, None, None
        for put_id, put_data in self.put_instructions.items():
            for col_name, col_data in put_data.items():
                if col_data.type in LayerOutputTypeChoice:
                    output_type = col_data.type
                    break

        if output_type == LayerOutputTypeChoice.Classification:
            print("\033[1mСписок классов и количество примеров:\033[0m")
            classes_names = self.put_instructions[put_id][col_name].parameters.classes_names
            dataframe_dict = {cl_name: [] for cl_name in classes_names}
            for split in ['train', 'val']:
                list_of_elements = self.dataframe[split].loc[:, col_name].tolist()
                for cl_name in classes_names:
                    dataframe_dict[cl_name].append(list_of_elements.count(cl_name))

            display(pd.DataFrame(dataframe_dict, index=['train', 'val']))

        elif output_type == LayerOutputTypeChoice.Segmentation:
            text_to_print = "\033[1mКлассы в масках сегментации и их цвета в RGB:\033[0m"
            classes = self.put_instructions[put_id][col_name].parameters.classes
            for name, color in classes.items():
                text_to_print += f"\n{name}: {color.as_rgb_tuple()}"
            print(text_to_print)

    def save(self, save_path: str) -> None:

        def arrays_save(arrays_data: Dict[str, Dict[int, np.ndarray]], path_to_folder: Path):
            for spl, data in arrays_data.items():
                for p_id, array in data.items():
                    joblib.dump(array, path_to_folder.joinpath(f"{p_id}_{spl}.gz"))

        path_to_save = Path(save_path)
        dataset_paths_data = utils.DatasetPathsData(path_to_save)

        arrays_save(self.X, dataset_paths_data.arrays.inputs)
        arrays_save(self.Y, dataset_paths_data.arrays.outputs)

        if self.preprocessing:
            for col_name, proc in self.preprocessing.items():
                if proc:
                    joblib.dump(proc, dataset_paths_data.preprocessing.joinpath(f"{col_name}.gz"))

        for split, dataframe in self.dataframe.items():
            dataframe.to_csv(dataset_paths_data.instructions.dataframe.joinpath(f"{split}.csv"))

        for put_id, cols_dict in self.put_instructions.items():
            for col_name, put_data in cols_dict.items():
                put_type = "input" if put_data.type in LayerInputTypeChoice else "output"
                file_name = f"{put_type}_{put_id}_{put_data.type}"
                with open(dataset_paths_data.instructions.parameters.joinpath(f"{file_name}.json"), "w") as instruction:
                    put_data.data = None
                    json.dump(put_data.json(), instruction)

        with open(dataset_paths_data.config, "w") as config:
            json.dump(self.dataset_data.json(), config)
        if not path_to_save.is_absolute():
            path_to_save = Path.cwd().joinpath(path_to_save)

        utils.logger.info(f"Датасет сохранен в директорию {path_to_save}")


class CreateDataset(TerraDataset):
    input_type: LayerInputTypeChoice = None
    output_type: LayerOutputTypeChoice = None

    def __init__(self, **kwargs):
        super().__init__()
        self.data = self._validate(
            getattr(dataset, f"{self.input_type}{self.output_type}Validator"), **kwargs
        )
        self.put_data = self.preprocess_put_data(
            data=self.data, data_type=LayerSelectTypeChoice.table
            if self.input_type == LayerInputTypeChoice.Dataframe else LayerSelectTypeChoice.folder
        )
        self.put_instructions = self.create_put_instructions(put_data=self.put_data)
        self.augmentation = self.create_augmentation(self.put_instructions)
        self.dataframe = self.create_table(
            self.put_instructions,
            train_size=self.data.train_size,
            shuffle=True if self.input_type != LayerInputTypeChoice.Timeseries else False,
            aug_coef=self.data.augmentation_coef,
            aug_data=self.augmentation
        )

        self.dataset_data = DatasetData(
            task=self.input_type.value + self.output_type.value,
            use_generator=False,
            is_created=False,
        )

        utils.logger.info(f"Датасет подготовлен к началу формирования массивов")

    @staticmethod
    def _validate(instance, **kwargs):
        data = instance(**kwargs)
        return data

    def preprocess_put_data(self, data, data_type: LayerSelectTypeChoice) -> \
            Dict[int, Union[Dict[Any, Any], Dict[str, InputData], Dict[str, OutputData]]]:

        puts_data = {}

        if data_type == LayerSelectTypeChoice.table:
            for idx, put in enumerate(data.inputs, 1):
                puts_data[idx] = {}
                for col_name in put.columns:
                    parameters_to_pass = {"csv_path": data.csv_path,
                                          "column": col_name,
                                          "type": put.type,
                                          "parameters": put.parameters}
                    put_data = InputData(**parameters_to_pass) if put.type in LayerInputTypeChoice else OutputData(
                        **parameters_to_pass)
                    puts_data[idx][f"{idx}_{col_name}"] = put_data
            puts_data[len(puts_data) + 1] = {f"{len(puts_data) + 1}_{data.output}": OutputData(
                csv_path=data.csv_path, column=data.output, type=self.output_type, parameters=
                getattr(outputs, f"{self.output_type}Validator")(**data.dict())
            )}

        elif data_type == LayerSelectTypeChoice.folder:
            puts_data[1] = {f"1_{self.input_type.value}": InputData(
                folder_path=process_directory_paths(data.source_path),
                column=f"1_{self.input_type.value}",
                type=self.input_type,
                parameters=getattr(inputs, f"{self.input_type.value}Validator")(**data.dict())
            )}
            puts_data[2] = {f"2_{self.output_type.value}": OutputData(
                folder_path=process_directory_paths(data.target_path) if "target_path" in data.__fields_set__
                else process_directory_paths(data.source_path),
                column=f"2_{self.output_type.value}",
                type=self.output_type,
                parameters=getattr(outputs, f"{self.output_type.value}Validator")(**data.dict())
            )}

        return puts_data

    @staticmethod
    def create_put_instructions(put_data) -> Dict[int, Dict[str, Union[InputInstructionsData, OutputInstructionsData]]]:

        new_put_data = {}
        for put_id, cols_dict in put_data.items():
            new_put_data[put_id] = {}
            for col_name, put_data in cols_dict.items():
                data_to_pass = []
                if put_data.csv_path:
                    csv_table = pd.read_csv(put_data.csv_path, usecols=[put_data.column])
                    data_to_pass = csv_table.loc[:, put_data.column].tolist()
                    if put_data.type in [LayerInputTypeChoice.Image, LayerOutputTypeChoice.Segmentation]:
                        data_to_pass = [str(put_data.csv_path.parent.joinpath(elem)) for elem in data_to_pass]
                    elif put_data.type in [LayerInputTypeChoice.Categorical, LayerOutputTypeChoice.Classification]:
                        put_data.parameters.classes_names = list(set(data_to_pass))
                else:
                    for folder_path in put_data.folder_path:
                        data_to_pass.extend(
                            getattr(utils, f"extract_{put_data.type.value.lower()}_data")(folder_path,
                                                                                          put_data.parameters)
                        )

                if put_data.type == LayerInputTypeChoice.Text:
                    if put_data.parameters.pymorphy:
                        data_to_pass = apply_pymorphy(data_to_pass)

                put_type = "Input" if put_data.type in LayerInputTypeChoice else "Output"
                parameters = put_data.parameters
                if put_data.type == LayerInputTypeChoice.Categorical:
                    parameters = CategoricalValidator(**put_data.parameters.dict())
                new_put_data[put_id][col_name] = getattr(creation_data, f"{put_type}InstructionsData")(
                    type=put_data.type, parameters=parameters, data=data_to_pass
                )

        return new_put_data

    @staticmethod
    def create_augmentation(put_instructions: Dict[int, Dict[str, Union[InputInstructionsData, OutputInstructionsData]]]):
        augmentation_data = {}
        for put_id, cols_dict in put_instructions.items():
            for col_name, put_data in cols_dict.items():
                if "augmentation" in put_data.parameters.dict() and put_data.parameters.augmentation:
                    aug_data = AugmentationData(aug_type=put_data.type, mode=put_data.parameters.augmentation)
                    augmentation_data[col_name] = \
                        getattr(augmentation, f"create_{put_data.type.lower()}_augmentation")(aug_data)
        return augmentation_data

    @staticmethod
    def create_table(put_instructions: Dict[int, Dict[str, Union[InputInstructionsData, OutputInstructionsData]]],
                     train_size: int,
                     aug_data: dict,
                     aug_coef: float,
                     shuffle: bool = True,
                     ) -> Dict[str, pd.DataFrame]:

        csv_data = {}
        for put_id, cols_dict in put_instructions.items():
            for col_name, put_data in cols_dict.items():
                csv_data[col_name] = put_data.data

        dataframe = pd.DataFrame.from_dict(csv_data)

        if shuffle:
            dataframe = dataframe.sample(frac=1)

        train_dataframe, val_dataframe = np.split(
            dataframe, [int(train_size * len(dataframe))]
        )

        if aug_data:
            for col_name, augm_data in aug_data.items():
                augm_df = train_dataframe.sample(int(len(train_dataframe) * aug_coef),
                                                 replace=False if aug_coef < 1 else True)
                augm_df[col_name] = augm_df[col_name].astype('str') + ";augm"
                train_dataframe = pd.concat([train_dataframe, augm_df], axis=0)
                train_dataframe = train_dataframe.sample(frac=1)

        dataframe = {"train": train_dataframe.reset_index(drop=True), "val": val_dataframe.reset_index(drop=True)}

        return dataframe


class CreateClassificationDataset(CreateDataset):

    def __init__(self, **kwargs):
        self.y_classes = []
        super().__init__(**kwargs)

    def balance_classes(self, balance_type: str) -> NoReturn:
        """
        В качестве аргумента option задаем способ балансирования классов
        Один из способов должен заключаться в копировании данных для уравнивания классов.
        Таким образом, массивы в генераторе будут содержать одинаковое количество классов.
        Другой способ - аугментация (при классификации изображений) с режимом Light
        - copy
        - augmentation
        :return:
        """

        if self.X["train"] and self.X["val"]:
            return "Балансирование классов возможна только перед вызовом метода .create(use_generator: bool = False)"

        dataframe = self.dataframe['train'].copy()
        image_col, cls_col = None, None
        for put_id, put_data in self.put_instructions.items():
            for col_name, col_data in put_data.items():
                if col_data.type == LayerInputTypeChoice.Image:
                    image_col = col_name
                elif col_data.type == LayerOutputTypeChoice.Classification:
                    cls_col = col_name
        if balance_type == "upsampling":
            classes_list = dataframe[cls_col].unique().tolist()
            classes_length = [len(dataframe[dataframe[cls_col] == cl_name])
                              for cl_name in classes_list]
            max_length = max(classes_length)
            df_list = []
            for cl_name in classes_list:
                df_list.append(resample(
                    dataframe[dataframe[cls_col] == cl_name],
                    replace=True,
                    n_samples=max_length)
                )
            self.dataframe['train'] = pd.concat(df_list).sample(frac=1).reset_index(drop=True)
            print(f"Максимальное количество примеров одного класса - {max_length} ({classes_list[classes_length.index(max_length)]}).")
            for cl_name, cl_length in zip(classes_list, classes_length):
                print(f"{cl_name}: {cl_length} - {max_length}")

        elif balance_type == "undersampling":
            dataframe = dataframe.groupby(cls_col)
            classes_list = dataframe.size().index.tolist()
            classes_length = dataframe.size().tolist()
            min_length = min(classes_length)
            self.dataframe['train'] = dataframe.apply(lambda x: x.sample(min_length))\
                .sample(frac=1).reset_index(drop=True)
            print(f"Минимальное количество примеров одного класса - {min_length} ({classes_list[classes_length.index(min_length)]}).")
            for cl_name, cl_length in zip(classes_list, classes_length):
                print(f"{cl_name}: {cl_length} - {min_length}")

        elif balance_type == "augmentation":
            if image_col:
                aug_data = AugmentationData(aug_type="Image", mode="Maximum")
                self.augmentation[image_col] = create_image_augmentation(image_augmentation_data=aug_data)
                classes_list = dataframe[cls_col].unique().tolist()
                classes_length = [len(dataframe[dataframe[cls_col] == cl_name])
                                  for cl_name in classes_list]
                max_length = max(classes_length)
                df_list = []
                for cl_name in classes_list:
                    data = dataframe[dataframe[cls_col] == cl_name]
                    cl_samples = classes_length[classes_list.index(cl_name)]
                    samples_to_add = max_length - cl_samples
                    if samples_to_add:
                        add_data = data.sample(samples_to_add)
                        add_data[image_col][~add_data[image_col].str.contains("augm")] = \
                            add_data[image_col][~add_data[image_col].str.contains("augm")] + ";augm"
                        data = pd.concat([data, add_data])
                    df_list.append(data)
                self.dataframe['train'] = pd.concat(df_list).sample(frac=1).reset_index(drop=True)
                print(f"Максимальное количество примеров одного класса - {max_length} ({classes_list[classes_length.index(max_length)]}).")
                for cl_name, cl_length in zip(classes_list, classes_length):
                    print(f"{cl_name}: {cl_length} - {max_length}")
            else:
                print('Способ балансировки классов "augmentation" доступен только для изображений')

        else:
            raise ValueError(f"Неизвестный способ балансировки классов: {balance_type}. "
                             f"Доступны: 'undersampling', 'upsampling', 'augmentation'")

    def create_put_instructions(self, put_data) -> \
            Dict[int, Dict[str, Union[InputInstructionsData, OutputInstructionsData]]]:

        new_put_data = {}
        for put_id, cols_dict in put_data.items():
            new_put_data[put_id] = {}
            for col_name, put_data in cols_dict.items():
                if put_data.type in LayerInputTypeChoice:
                    data_to_pass = []
                    for folder_path in put_data.folder_path:
                        data = getattr(utils, f"extract_{put_data.type.value.lower()}_data")(
                            folder_path, put_data.parameters
                        )
                        if put_id == 1:
                            self.y_classes.extend([folder_path.name for _ in data])
                        data_to_pass.extend([str(path) for path in data])
                    if put_data.type == LayerInputTypeChoice.Text:
                        if put_data.parameters.pymorphy:
                            data_to_pass = apply_pymorphy(data_to_pass)
                else:
                    data_to_pass = self.y_classes
                    put_data.parameters.classes_names = [path.name for path in put_data.folder_path]

                put_type = "Input" if put_data.type in LayerInputTypeChoice else "Output"
                new_put_data[put_id][col_name] = getattr(creation_data, f"{put_type}InstructionsData")(
                    type=put_data.type, parameters=put_data.parameters, data=data_to_pass
                )

        return new_put_data



import errno
import os
from pathlib import Path
from typing import List

from pydantic import BaseModel, validator
from pydantic.types import PositiveFloat

from terra_ai_datasets.creation.validators import inputs, outputs
from terra_ai_datasets.creation.validators.tasks import LayerInputTypeChoice, LayerOutputTypeChoice


# --- Common validators ---
class CommonValidator(BaseModel):
    train_size: float
    augmentation_coef: float = 0.0

    @validator('train_size')
    def train_size_range(cls, v):
        if not 0.0 < v <= 1.0:
            raise ValueError('Ensure "train_size" is in range between 0.0 and 1.0')
        return v

    @validator('augmentation_coef')
    def train_size_range(cls, v):
        if not 0.0 <= v:
            raise ValueError('Ensure "augmentation_coef" is greater than 0')
        return v

    class Config:
        use_enum_values = True


class SourceFolderPathValidator(CommonValidator):
    source_path: List[Path]

    @validator('source_path', each_item=True)
    def sources_folder_exists(cls, source_path: Path):
        if not source_path.is_dir():
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), source_path
                )
        return source_path


class TargetFolderPathValidator(CommonValidator):
    target_path: List[Path]

    @validator('target_path', each_item=True)
    def target_folder_exists(cls, target_path: Path):
        if not target_path.is_dir():
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), target_path
                )
        return target_path


class FilePathValidator(CommonValidator):
    csv_path: Path

    @validator('csv_path')
    def file_path_exists(cls, csv_path: Path):
        if not csv_path.is_file():
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), csv_path
                )
        if not csv_path.suffix == '.csv':
            raise Exception('Must be a .csv file')
        return csv_path


# --- Image validators ---
class ImageClassificationValidator(SourceFolderPathValidator, inputs.ImageValidator, outputs.ClassificationValidator):
    pass


class ImageSegmentationValidator(SourceFolderPathValidator, TargetFolderPathValidator,
                                 inputs.ImageValidator, outputs.SegmentationValidator):
    pass


# --- Text validators ---
class TextClassificationValidator(SourceFolderPathValidator, inputs.TextValidator, outputs.ClassificationValidator):
    pass


# --- Audio validators ---
class AudioClassificationValidator(SourceFolderPathValidator, inputs.AudioValidator, outputs.ClassificationValidator):
    pass


# --- Dataframe Validators ---
class DataframePutData(BaseModel):
    columns: List[str]


class DataframeInputData(DataframePutData):
    type: LayerInputTypeChoice
    parameters: dict

    @validator("parameters")
    def validate_input_parameters(cls, parameters, values):
        if not values.get("type"):
            raise ValueError
        return getattr(inputs, f'{values["type"].value}Validator')(**parameters)


class DataframeOutputData(DataframePutData):
    type: LayerOutputTypeChoice
    parameters: dict

    @validator("parameters")
    def validate_output_parameters(cls, parameters, values):
        if not values.get("type"):
            raise ValueError
        return getattr(outputs, f'{values["type"].value}Validator')(**parameters)


class DataframeValidator(BaseModel):
    inputs: List[DataframeInputData]
    output: str


class DataframeClassificationValidator(FilePathValidator, DataframeValidator, outputs.ClassificationValidator):
    pass


class DataframeRegressionValidator(FilePathValidator, DataframeValidator, outputs.RegressionValidator):
    pass


class TimeseriesDepthValidator(FilePathValidator, inputs.TimeseriesValidator, outputs.DepthValidator):
    inputs: List[str]
    outputs: List[str]


class TimeseriesTrendValidator(FilePathValidator, inputs.TimeseriesValidator, outputs.TrendValidator):
    inputs: List[str]
    output: str



from datetime import datetime

from pydantic import BaseModel

from terra_ai_datasets.creation.validators.tasks import TasksChoice


class DatasetData(BaseModel):
    task: TasksChoice
    use_generator: bool
    is_created: bool
    date: datetime = datetime.now().isoformat()

    class Config:
        use_enum_values = True



from typing import Optional, List, Any, Dict

from pydantic import BaseModel, PositiveInt, DirectoryPath, FilePath

from terra_ai_datasets.creation.validators.tasks import LayerInputTypeChoice, LayerOutputTypeChoice


# class CSVData(BaseModel):
#     csv_path: FilePath
#     columns: List[str]


class PutData(BaseModel):
    csv_path: Optional[FilePath]
    column: Optional[str]
    folder_path: Optional[List[DirectoryPath]]
    parameters: Any


class InputData(PutData):
    type: LayerInputTypeChoice

    # @validator("parameters")
    # def validate_input_parameters(cls, parameters, values):
    #     return getattr(inputs, f'{values["type"].value}Validator')(**parameters)


class OutputData(PutData):
    type: LayerOutputTypeChoice

    # @validator("parameters")
    # def validate_output_parameters(cls, parameters, values):
    #     return getattr(outputs, f'{values["type"].value}Validator')(**parameters)


class BaseInstructionsData(BaseModel):
    parameters: Any
    data: Optional[List[Any]]


class InputInstructionsData(BaseInstructionsData):
    type: LayerInputTypeChoice


class OutputInstructionsData(BaseInstructionsData):
    type: LayerOutputTypeChoice


class DatasetInstructionsData(BaseModel):
    inputs: Dict[PositiveInt, InputInstructionsData]
    outputs: Dict[PositiveInt, OutputInstructionsData]



import imgaug.augmenters as iaa

from terra_ai_datasets.creation.validators.inputs import AugmentationData


def create_image_augmentation(image_augmentation_data: AugmentationData):
    parameters = []
    for key, value in image_augmentation_data.parameters.dict().items():
        if value:
            parameters.append(getattr(iaa, key)(**value))
    seq = iaa.Sequential(parameters, random_order=True)
    return seq
